r"""Stable Diffusion (SD) plugin.

This plugin depends on :mod:`diffusers` and :mod:`transformers`. To use it,
install the dependencies in your environment

.. code-block:: console

    pip install diffusers transformers accelerate

before importing the plugin.

.. code-block:: python

    from azula.plugins import sd

References:
    | High-Resolution Image Synthesis with Latent Diffusion Models (Rombach et al., 2021)
    | https://arxiv.org/abs/2112.10752
"""

__all__ = [
    "AutoEncoder",
    "TextEncoder",
    "StableDenoiser",
    "load_model",
]

import torch
import torch.nn as nn

from collections.abc import Sequence
from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import Schedule, VPSchedule

from ..utils import load_cards, patch_diffusers, patch_mmap_safetensors


class AutoEncoder(nn.Module):
    r"""Creates an auto-encoder wrapper."""

    def __init__(
        self,
        vae: nn.Module,  # AutoencoderKL
        scale: float = 1.0,
    ) -> None:
        super().__init__()

        self.vae = vae
        self.scale = scale

    def encode(self, x: Tensor) -> Tensor:
        r"""Encodes images to latents.

        Arguments:
            x: A batch of images :math:`x`, with shape :math:`(B, 3, H, W)`.
                Pixel values are expected to range between 0 and 1.

        Returns:
            A batch of latents :math:`z \sim q(Z \mid x)`, with shape :math:`(B, 4, H / 8, W / 8)`.
        """

        dtype = get_module_dtype(self.vae)

        q_z_x = self.vae.encode(x.to(dtype)).latent_dist
        z = q_z_x.mean + q_z_x.std * torch.randn_like(q_z_x.mean)
        z = z * self.scale

        return z.to(x)

    def decode(self, z: Tensor) -> Tensor:
        r"""Decodes latents to images.

        Arguments:
            z: A batch of latents :math:`z`, with shape :math:`(B, 4, H / 8, W / 8)`.

        Returns:
            A batch of images :math:`x = D(z)`, with shape :math:`(B, 3, H, W)`.
        """

        dtype = get_module_dtype(self.vae)

        z = z / self.scale
        x = self.vae.decode(z.to(dtype)).sample

        return x.to(z)


class TextEncoder(nn.Module):
    r"""Creates a text encoder."""

    def __init__(
        self,
        clip: nn.Module,
        tokenizer: nn.Module,
    ) -> None:
        super().__init__()

        self.clip = clip
        self.tokenizer = tokenizer

    def forward(self, prompt: str | Sequence[str]) -> dict[str, Tensor]:
        r"""
        Arguments:
            prompt: A text prompt or list of text prompts.

        Returns:
            The CLIP encoded prompt(s).
        """

        if isinstance(prompt, str):
            prompt = [prompt]

        tokens = self.tokenizer(
            prompt,
            truncation=True,
            max_length=self.tokenizer.model_max_length,
            padding="max_length",
            return_tensors="pt",
        )

        if getattr(self.clip.config, "use_attention_mask", False):
            mask = tokens.attention_mask
        else:
            mask = None

        embeds = self.clip(
            input_ids=tokens.input_ids,
            attention_mask=mask,
            output_hidden_states=False,
        ).last_hidden_state

        return {
            "prompt_embeds": embeds,
        }


class StableDenoiser(Denoiser):
    r"""Creates a stable denoiser.

    Arguments:
        backbone: A time conditional network.
        sigmas: The discrete noise schedule used during training.
        schedule: A noise schedule. If :py:`None`, use
            :class:`azula.noise.VPSchedule` instead.
        prediction: The backbone prediction type.
    """

    def __init__(
        self,
        backbone: nn.Module,
        sigmas: Tensor,
        schedule: Schedule | None = None,
        prediction: str = "epsilon",
    ) -> None:
        super().__init__()

        self.backbone = backbone
        self.prediction = prediction

        if schedule is None:
            self.schedule = VPSchedule(
                alpha_min=(1 - sigmas[-1].item() ** 2) ** 0.5,
                sigma_min=sigmas[0].item(),
            )
        else:
            self.schedule = schedule

        self.register_buffer("sigmas", sigmas.to(torch.get_default_dtype()))

    def forward(
        self,
        z_t: Tensor,
        t: Tensor,
        prompt_embeds: Tensor,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            z_t: A noisy tensor :math:`z_t`, with shape :math:`(B, C, H, W)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            prompt_embeds: The CLIP-encoded text prompt :math:`y`, with shape :math:`(B, L, D)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Gaussian :math:`\mathcal{N}(Z \mid \mu_\phi(z_t \mid y), \Sigma_\phi(z_t \mid y)`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < z_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        if self.prediction == "epsilon":
            c_out = -sigma_t / alpha_t
            c_skip = 1 / alpha_t
        elif self.prediction == "velocity":
            c_out = -sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
            c_skip = alpha_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        else:
            raise ValueError(f"Unkown prediction type '{self.prediction}'.")

        c_in = torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_time = sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_time = torch.searchsorted(self.sigmas, c_time.flatten())

        B, _, _, _ = z_t.shape
        _, L, D = prompt_embeds.shape

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            timestep=c_time.expand(B),
            sample=(c_in * z_t).to(dtype),
            encoder_hidden_states=prompt_embeds.to(dtype).expand(B, L, D),
            **kwargs,
        ).sample.to(z_t)

        mean = c_skip * z_t + c_out * output

        return DiracPosterior(mean=mean)


def load_model(
    name: str,
    **kwargs,
) -> tuple[Denoiser, AutoEncoder, TextEncoder]:
    r"""Loads a pre-trained stable latent denoiser.

    Arguments:
        name: The pre-trained model name.
        kwargs: Keyword arguments passed to :func:`diffusers.StableDiffusionPipeline.from_pretrained`.

    Returns:
        A pre-trained latent denoiser and the corresponding auto-encoder and text encoder.
    """

    from diffusers import StableDiffusionPipeline

    card = load_cards(__name__)[name]

    with skip_init(), patch_diffusers():
        pipe = StableDiffusionPipeline.from_pretrained(
            card.repo,
            variant=card.variant,
            torch_dtype=card.dtype_map,
            **kwargs,
        )

    patch_mmap_safetensors(pipe.unet, pipe.vae, pipe.text_encoder)

    alphas = pipe.scheduler.alphas_cumprod.to(dtype=torch.float64).sqrt()
    sigmas = torch.sqrt(1 - alphas**2)

    denoiser = StableDenoiser(
        backbone=pipe.unet,
        sigmas=sigmas,
        **card.config,
    )

    autoencoder = AutoEncoder(
        vae=pipe.vae,
        scale=pipe.vae.config.scaling_factor,
    )

    textencoder = TextEncoder(
        clip=pipe.text_encoder,
        tokenizer=pipe.tokenizer,
    )

    return denoiser.eval(), autoencoder.eval(), textencoder.eval()



r"""Diffusion Plug-and-Play Image Restoration (DiffPIR) internals.

References:
    | Denoising Diffusion Models for Plug-and-Play Image Restoration (Zhu et al., 2023)
    | https://arxiv.org/abs/2305.08995
"""

__all__ = [
    "DiffPIRDenoiser",
]

import torch

from collections.abc import Callable
from functools import partial
from torch import Tensor
from typing import Literal

from ..denoise import Denoiser, DiracPosterior
from ..linalg.solve import cg, gmres
from ..noise import Schedule


class DiffPIRDenoiser(Denoiser):
    r"""Creates a DiffPIR denoiser module.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y \sim \mathcal{N}(A x, \Sigma_y)`, with shape :math:`(*, D)`.
        A: The forward operator :math:`x \mapsto A x`.
        var_y: The noise variance :math:`\Sigma_y`.
        lmbda: The regularization strength :math:`\lambda \in \mathbb{R}_+`.
        solver: The linear solver name.
        iterations: The number of solver iterations.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        A: Callable[[Tensor], Tensor],
        var_y: float | Tensor,
        lmbda: float = 10.0,
        solver: Literal["cg", "gmres"] = "gmres",
        iterations: int = 1,
    ) -> None:
        super().__init__()

        self.denoiser = denoiser

        self.y = y
        self.A = A
        self.var_y = var_y
        self.lmbda = lmbda

        if solver == "cg":
            self.solve = partial(cg, iterations=iterations)
        elif solver == "gmres":
            self.solve = partial(gmres, iterations=iterations)
        else:
            raise ValueError(f"Unknown solver '{solver}'.")

    @property
    def schedule(self) -> Schedule:
        return self.denoiser.schedule

    @torch.no_grad()
    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        alpha_t, sigma_t = self.schedule(t)
        rho_t = (sigma_t / alpha_t) ** 2

        q = self.denoiser(x_t, t, **kwargs)

        with torch.enable_grad():
            x_hat = q.mean.detach().requires_grad_()
            y_hat = self.A(x_hat)

        def At(v: Tensor) -> Tensor:
            return torch.autograd.grad(y_hat, x_hat, v, retain_graph=True)[0]

        def AtA_I(v: Tensor) -> Tensor:
            return At(self.A(v) / self.var_y) + self.lmbda * v / rho_t

        grad = (self.y - y_hat) / self.var_y
        grad = At(grad)
        grad = self.solve(A=AtA_I, b=grad)

        return DiracPosterior(mean=x_hat + grad)



r"""Flux plugin.

This plugin depends on :mod:`diffusers` and :mod:`transformers`. To use it,
install the dependencies in your environment

.. code-block:: console

    pip install diffusers transformers accelerate protobuf sentencepiece

before importing the plugin.

.. code-block:: python

    from azula.plugins import flux

References:
    | FLUX (Black Forest Labs, 2024)
    | https://github.com/black-forest-labs/flux
"""

__all__ = [
    "AutoEncoder",
    "TextEncoder",
    "FluxDenoiser",
    "load_model",
]

import torch
import torch.nn as nn

from collections.abc import Sequence
from einops import rearrange
from functools import cache
from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import DecaySchedule, Schedule

from ..utils import load_cards, patch_diffusers, patch_mmap_safetensors


class AutoEncoder(nn.Module):
    r"""Creates an auto-encoder wrapper."""

    def __init__(
        self,
        vae: nn.Module,  # AutoencoderKL
        shift: float = 0.0,
        scale: float = 1.0,
    ) -> None:
        super().__init__()

        self.vae = vae
        self.shift = shift
        self.scale = scale

    def encode(self, x: Tensor) -> Tensor:
        r"""Encodes images to latents.

        Arguments:
            x: A batch of images :math:`x`, with shape :math:`(B, 3, H, W)`.
                Pixel values are expected to range between -1 and 1.

        Returns:
            A batch of latents :math:`z \sim q(Z \mid x)`, with shape :math:`(B, H / 16, W / 16, 64)`.
        """

        dtype = get_module_dtype(self.vae)

        q_z_x = self.vae.encode(x.to(dtype)).latent_dist
        z = q_z_x.mean + q_z_x.std * torch.randn_like(q_z_x.mean)
        z = (z - self.shift) * self.scale
        z = rearrange(z, "... C (H h) (W w) -> ... H W (C h w)", h=2, w=2)

        return z.to(x)

    def decode(self, z: Tensor) -> Tensor:
        r"""Decodes latents to images.

        Arguments:
            z: A batch of latents :math:`z`, with shape :math:`(B, H / 16, W / 16, 64)`.

        Returns:
            A batch of images :math:`x = D(z)`, with shape :math:`(B, 3, H, W)`.
        """

        dtype = get_module_dtype(self.vae)

        z = rearrange(z, "... H W (C h w) -> ... C (H h) (W w)", h=2, w=2)
        z = z / self.scale + self.shift
        x = self.vae.decode(z.to(dtype)).sample

        return x.to(z)


class TextEncoder(nn.Module):
    r"""Creates a text encoder."""

    def __init__(
        self,
        clip: nn.Module,
        clip_tokenizer: nn.Module,
        t5: nn.Module,
        t5_tokenizer: nn.Module,
    ) -> None:
        super().__init__()

        self.clip = clip
        self.clip_tokenizer = clip_tokenizer
        self.t5 = t5
        self.t5_tokenizer = t5_tokenizer

    def forward(self, prompt: str | Sequence[str]) -> dict[str, Tensor]:
        r"""
        Arguments:
            prompt: A text prompt or list of text prompts.

        Returns:
            The CLIP and T5 encoded prompt(s).
        """

        if isinstance(prompt, str):
            prompt = [prompt]

        # CLIP
        clip_output = self.clip(
            input_ids=self.clip_tokenizer(
                prompt,
                truncation=True,
                max_length=self.clip_tokenizer.model_max_length,
                padding="max_length",
                return_tensors="pt",
            ).input_ids,
            output_hidden_states=False,
        ).pooler_output

        # T5
        t5_output = self.t5(
            input_ids=self.t5_tokenizer(
                prompt,
                truncation=True,
                max_length=self.t5_tokenizer.model_max_length,
                padding="max_length",
                return_tensors="pt",
            ).input_ids,
            output_hidden_states=False,
        ).last_hidden_state

        return {
            "prompt_clip": clip_output,
            "prompt_t5": t5_output,
        }


class FluxDenoiser(Denoiser):
    r"""Creates a Flux denoiser.

    Arguments:
        backbone: A time conditional network.
        schedule: A noise schedule. If :py:`None`, use
            :class:`azula.noise.DecaySchedule` instead.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = DecaySchedule()
        else:
            self.schedule = schedule

    @staticmethod
    @cache
    def coordinates(
        H: int,
        W: int,
        dtype: torch.dtype,
        device: torch.device,
    ) -> Tensor:
        return torch.stack(
            torch.meshgrid(
                torch.zeros(1, dtype=dtype, device=device),
                torch.arange(H, dtype=dtype, device=device),
                torch.arange(W, dtype=dtype, device=device),
                indexing="ij",
            ),
            dim=-1,
        ).reshape(-1, 3)

    def forward(
        self,
        z_t: Tensor,
        t: Tensor,
        prompt_clip: Tensor,
        prompt_t5: Tensor,
        guidance: float | Tensor = 4.0,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            z_t: A noisy tensor :math:`z_t`, with shape :math:`(B, H, W, 64)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            prompt_clip: The CLIP-encoded text prompt :math:`y`, with shape :math:`(B, F)`.
            prompt_t5: The T5-encoded text prompt :math:`y`, with shape :math:`(B, L, D)`.
            guidance: The guidance strength :math:`\omega \in \mathbb{R}`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(Z - \mu_\phi(z_t \mid y))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < z_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = 1 / (alpha_t + sigma_t)
        c_out = -sigma_t / (alpha_t + sigma_t)
        c_skip = 1 / (alpha_t + sigma_t)
        c_time = (sigma_t / (alpha_t + sigma_t)).flatten()

        B, H, W, C = z_t.shape
        _, L, D = prompt_t5.shape

        dtype = get_module_dtype(self.backbone)

        img_ids = self.coordinates(H, W, dtype=dtype, device=z_t.device)
        txt_ids = torch.zeros((L, 3), dtype=dtype, device=z_t.device)

        if guidance is not None:
            guidance = torch.as_tensor(guidance, dtype=dtype, device=z_t.device).expand(B)

        output = (
            self.backbone(
                timestep=c_time.to(dtype).expand(B),
                hidden_states=(c_in * z_t).to(dtype).reshape(B, H * W, C),
                encoder_hidden_states=prompt_t5.to(dtype).expand(B, L, D),
                pooled_projections=prompt_clip.to(dtype),
                img_ids=img_ids,
                txt_ids=txt_ids,
                guidance=guidance,
                **kwargs,
            )
            .sample.reshape_as(z_t)
            .to(z_t)
        )

        mean = c_skip * z_t + c_out * output

        return DiracPosterior(mean=mean)


def load_model(
    name: str = "flux_1_dev",
    **kwargs,
) -> tuple[Denoiser, AutoEncoder, TextEncoder]:
    r"""Loads a pre-trained Flux latent denoiser.

    Arguments:
        name: The pre-trained model name.
        kwargs: Keyword arguments passed to :func:`diffusers.FluxPipeline.from_pretrained`.

    Returns:
        A pre-trained latent denoiser and the corresponding auto-encoder and text encoder.
    """

    from diffusers import FluxPipeline

    card = load_cards(__name__)[name]

    with skip_init(), patch_diffusers():
        pipe = FluxPipeline.from_pretrained(
            card.repo,
            variant=card.variant,
            torch_dtype=card.dtype_map,
            **kwargs,
        )

    patch_mmap_safetensors(pipe.transformer, pipe.vae, pipe.text_encoder, pipe.text_encoder_2)

    denoiser = FluxDenoiser(backbone=pipe.transformer)

    autoencoder = AutoEncoder(
        vae=pipe.vae,
        shift=pipe.vae.config.shift_factor,
        scale=pipe.vae.config.scaling_factor,
    )

    textencoder = TextEncoder(
        clip=pipe.text_encoder,
        clip_tokenizer=pipe.tokenizer,
        t5=pipe.text_encoder_2,
        t5_tokenizer=pipe.tokenizer_2,
    )

    return denoiser.eval(), autoencoder.eval(), textencoder.eval()



r"""U-Net building blocks."""

__all__ = [
    "UNet",
    "UNetBlock",
]

import torch

from collections.abc import Sequence
from einops.layers.torch import Rearrange
from torch import Tensor

from .layers import ConvNd, LayerNorm, RMSNorm
from .utils import checkpoint


class UNetBlock(torch.nn.Module):
    r"""Creates a modulated U-Net block module.

    Arguments:
        channels: The number of channels :math:`C`.
        mod_features: The number of modulating features :math:`D`.
        norm: The kind of normalization. Options are `group`, `layer` and `rms`.
        groups: The number of groups in :class:`torch.nn.GroupNorm` layers.
        ffn_factor: The channel factor in the FFN.
        spatial: The number of spatial dimensions :math:`N`.
        dropout: The dropout rate in :math:`[0, 1]`.
        checkpointing: Whether to use activation checkpointing or not.
        kwargs: Keyword arguments passed to :class:`azula.nn.layers.ConvNd`.
    """

    def __init__(
        self,
        channels: int,
        mod_features: int = 0,
        norm: str = "layer",
        groups: int = 16,
        ffn_factor: int = 1,
        spatial: int = 2,
        dropout: float | None = None,
        checkpointing: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()

        self.checkpointing = checkpointing

        # Ada-Norm Zero
        if norm == "layer":
            self.norm = LayerNorm(dim=-spatial - 1, eps=1e-5)
        elif norm == "rms":
            self.norm = RMSNorm(dim=-spatial - 1, eps=1e-5)
        elif norm == "group":
            self.norm = torch.nn.GroupNorm(
                num_groups=min(groups, channels),
                num_channels=channels,
                affine=False,
                eps=1e-5,
            )
        else:
            raise NotImplementedError()

        if mod_features > 0:
            self.ada_zero = torch.nn.Sequential(
                torch.nn.Linear(mod_features, mod_features),
                torch.nn.SiLU(),
                torch.nn.Linear(mod_features, 3 * channels),
                Rearrange("... (n C) -> n ... C" + " 1" * spatial, n=3),
            )

            self.ada_zero[-2].weight.data.mul_(1e-2)
        else:
            self.ada_zero = torch.nn.Parameter(torch.randn(3, channels, *(1,) * spatial))
            self.ada_zero.data.mul_(1e-2)

        # Block
        self.ffn = torch.nn.Sequential(
            ConvNd(channels, ffn_factor * channels, spatial=spatial, **kwargs),
            torch.nn.SiLU(),
            torch.nn.Identity() if dropout is None else torch.nn.Dropout(dropout),
            ConvNd(ffn_factor * channels, channels, spatial=spatial, **kwargs),
        )

    def _forward(self, x: Tensor, mod: Tensor | None = None) -> Tensor:
        if torch.is_tensor(self.ada_zero):
            a, b, c = self.ada_zero
        else:
            a, b, c = self.ada_zero(mod)

        y = (a + 1) * self.norm(x) + b
        y = self.ffn(y)
        y = x + c * y

        return y

    def forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tensor, with shape :math:`(B, C, L_1, ..., L_N)`.
            mod: The modulation vector, with shape :math:`(D)` or :math:`(B, D)`.

        Returns:
            The output tensor, with shape :math:`(B, C, L_1, ..., L_N)`.
        """

        if self.checkpointing:
            return checkpoint(self._forward, reentrant=not self.training)(x, mod)
        else:
            return self._forward(x, mod)


class UNet(torch.nn.Module):
    r"""Creates a modulated U-Net module.

    Arguments:
        in_channels: The number of input channels :math:`C_i`.
        out_channels: The number of output channels :math:`C_o`.
        cond_channels: The number of condition channels :math:`C_c`.
        hid_channels: The numbers of channels at each depth.
        hid_blocks: The numbers of hidden blocks at each depth.
        kernel_size: The kernel size of all convolutions.
        stride: The stride of the downsampling convolutions.
        spatial: The number of spatial dimensions :math:`N`.
        periodic: Whether the spatial dimensions are periodic or not.
        identity_init: Initialize down/upsampling convolutions as identity.
        kwargs: Keyword arguments passed to :class:`UNetBlock`.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        cond_channels: int = 0,
        hid_channels: Sequence[int] = (64, 128, 256),
        hid_blocks: Sequence[int] = (3, 3, 3),
        kernel_size: int | Sequence[int] = 3,
        stride: int | Sequence[int] = 2,
        spatial: int = 2,
        periodic: bool = False,
        identity_init: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()

        assert len(hid_blocks) == len(hid_channels)

        if isinstance(kernel_size, int):
            kernel_size = [kernel_size] * spatial

        if isinstance(stride, int):
            stride = [stride] * spatial

        conv_kwargs = dict(  # noqa: C408
            kernel_size=tuple(kernel_size),
            padding=tuple(k // 2 for k in kernel_size),
            padding_mode="circular" if periodic else "zeros",
            spatial=spatial,
        )

        self.descent, self.ascent = torch.nn.ModuleList(), torch.nn.ModuleList()

        for i, num_blocks in enumerate(hid_blocks):
            do, up = torch.nn.ModuleList(), torch.nn.ModuleList()

            for _ in range(num_blocks):
                do.append(UNetBlock(hid_channels[i], **conv_kwargs, **kwargs))
                up.append(UNetBlock(hid_channels[i], **conv_kwargs, **kwargs))

            if i > 0:
                do.insert(
                    0,
                    ConvNd(
                        hid_channels[i - 1],
                        hid_channels[i],
                        stride=stride,
                        identity_init=identity_init,
                        **conv_kwargs,
                    ),
                )

                up.append(torch.nn.Upsample(scale_factor=tuple(stride), mode="nearest"))
            else:
                do.insert(0, ConvNd(in_channels + cond_channels, hid_channels[i], **conv_kwargs))
                up.append(ConvNd(hid_channels[i], out_channels, **conv_kwargs))

            if i + 1 < len(hid_blocks):
                up.insert(
                    0,
                    ConvNd(
                        hid_channels[i] + hid_channels[i + 1],
                        hid_channels[i],
                        identity_init=identity_init,
                        **conv_kwargs,
                    ),
                )

            self.descent.append(do)
            self.ascent.insert(0, up)

    def forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
        cond: Tensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tensor, with shape :math:`(B, C_i, L_1, ..., L_N)`.
            mod: The modulation vector, with shape :math:`(D)` or :math:`(B, D)`.
            cond: The condition tensor, with shape :math:`(B, C_c, L_1, ..., L_N)`.

        Returns:
            The output tensor, with shape :math:`(B, C_o, L_1, ..., L_N)`.
        """

        if cond is not None:
            x = torch.cat((x, cond), dim=1)

        memory = []

        for blocks in self.descent:
            if memory:
                memory.append(x)
            else:
                memory.append(None)

            for block in blocks:
                if isinstance(block, UNetBlock):
                    x = block(x, mod)
                else:
                    x = block(x)

        if hasattr(self, "bottleneck"):
            x = self.bottleneck(x, mod)

        for blocks in self.ascent:
            for block in blocks:
                if isinstance(block, UNetBlock):
                    x = block(x, mod)
                else:
                    x = block(x)

            y = memory.pop()

            if y is None:
                continue

            for i in range(2, x.ndim):
                if x.shape[i] > y.shape[i]:
                    x = torch.narrow(x, i, 0, y.shape[i])

            x = torch.cat((y, x), dim=1)

        return x



r"""Miscellaneous neural network helpers."""

__all__ = [
    "checkpoint",
    "get_module_device",
    "get_module_dtype",
    "promote_dtype",
    "skip_init",
]

import torch
import torch.utils.checkpoint

from collections.abc import Callable
from functools import reduce, wraps
from torch import Tensor
from torch._C._functorch import is_gradtrackingtensor
from torch.autograd.function import FunctionCtx
from typing import TypeVar

T = TypeVar("T")


def get_module_dtype(module: torch.nn.Module) -> torch.dtype:
    r"""Returns the data type of a module.

    The module's data type is the first floating-point type in the module's parameters
    or buffers. If there is none, returns :py:`None`.

    Arguments:
        module: A module.
    """

    for p in module.parameters():
        if torch.is_floating_point(p):
            return p.dtype

    for b in module.buffers():
        if torch.is_floating_point(b):
            return b.dtype

    return None


def get_module_device(module: torch.nn.Module) -> torch.device:
    r"""Returns the execution device of a module.

    The module's device is the first device in the module's parameters or buffers. If
    there is none, returns :py:`None`.

    Arguments:
        module: A module.
    """

    for m in module.modules():
        if (
            hasattr(m, "_hf_hook")
            and hasattr(m._hf_hook, "execution_device")
            and m._hf_hook.execution_device is not None
        ):
            return m._hf_hook.execution_device

        for p in m.parameters(recurse=False):
            if not p.is_meta():
                return p.device

        for b in m.parameters(recurse=False):
            if not b.is_meta():
                return b.device

    return None


class CheckpointReentrant(torch.autograd.Function):
    @staticmethod
    def setup_context(
        ctx: FunctionCtx,
        inputs: tuple[Callable[..., Tensor | tuple[Tensor, ...]], Tensor],
        outputs: Tensor | tuple[Tensor, ...],
    ) -> None:
        func, *xs = inputs

        xs = [x.detach() for x in xs]

        ctx.save_for_backward(*xs)
        ctx.save_for_forward(*xs)
        ctx.func = func

    @staticmethod
    def forward(func: Callable[..., Tensor | tuple[Tensor, ...]], *xs: Tensor) -> T:
        return func(*xs)

    @staticmethod
    @torch.autograd.function.once_differentiable
    def vjp(ctx: FunctionCtx, *grad_ys: Tensor) -> tuple[None, Tensor]:
        xs = ctx.saved_tensors

        with torch.enable_grad():
            xs = [x.detach().requires_grad_() for x in xs]
            ys = ctx.func(*xs)

        if torch.is_tensor(ys):
            ys = [ys]

        grad_xs = torch.autograd.grad(ys, xs, grad_ys)

        return (None, *grad_xs)

    @staticmethod
    @torch.autograd.function.once_differentiable
    def jvp(ctx: FunctionCtx, grad_func: None, *grad_xs: Tensor) -> Tensor | tuple[Tensor, ...]:
        xs = ctx.saved_tensors

        _, grad_ys = torch.func.jvp(ctx.func, xs, grad_xs)

        return grad_ys


def checkpoint(f: Callable[..., T], reentrant: bool = False) -> Callable[..., T]:
    r"""Applies activation checkpointing to a function.

    Activation checkpointing reduces memory consumption by storing the inputs of the
    function and recomputing its graph during automatic differentiation (AD).

    Reentrant checkpointing is compatible with backward and forward AD, but only
    propagates gradients to the explicit positional inputs of the function. Implicit
    inputs, such as module parameters, do not get gradients. Conversely, non-reentrant
    will propagate gradients to implicit inputs, but is not compatible with foward AD.

    Arguments:
        f: A function.
        reentrant: Whether to use reentrant checkpointing or not.

    Returns:
        The checkpointed function.
    """

    @wraps(f)
    def g(*args, **kwargs) -> T:
        if not torch.is_grad_enabled():
            return f(*args, **kwargs)

        mask = [
            torch.is_tensor(arg)
            and torch.is_floating_point(arg)
            and (arg.requires_grad or is_gradtrackingtensor(arg))
            for arg in args
        ]

        tensors = [arg for include, arg in zip(mask, args, strict=True) if include]
        others = [arg for include, arg in zip(mask, args, strict=True) if not include]

        def h(*tensors: Tensor) -> T:
            it, io = iter(tensors), iter(others)
            args = (next(it if include else io) for include in mask)
            return f(*args, **kwargs)

        if any(map(is_gradtrackingtensor, tensors)):
            return h(*tensors)
        elif reentrant:
            if tensors:
                return CheckpointReentrant.apply(h, *tensors)
            else:
                with torch.no_grad():
                    return h(*tensors)
        else:
            return torch.utils.checkpoint.checkpoint(h, *tensors, use_reentrant=False)

    return g


class skip_init(torch.overrides.TorchFunctionMode):
    r"""Creates a context in which weight initialization is skipped.

    Example:
        >>> with skip_init():
        ...    layer = torch.nn.Linear(3, 5)
    """

    def __torch_function__(self, func, types, args=(), kwargs=None) -> Tensor:  # noqa: ANN001
        kwargs = kwargs or {}
        if getattr(func, "__module__", None) == "torch.nn.init":
            if "tensor" in kwargs:
                return kwargs["tensor"]
            else:
                return args[0]
        else:
            return func(*args, **kwargs)


def promote_dtype(f: Callable[..., T], min_dtype: torch.dtype = torch.float32) -> Callable[..., T]:
    r"""Applies data type promotion to a function.

    Arguments:
        f: A function.
        min_dtype: The minimum precision data type.

    Returns:
        The promoted function.
    """

    @wraps(f)
    def g(*args, **kwargs) -> T:
        if any(arg.device.type == "mps" for arg in args) and min_dtype is torch.float64:
            args = [arg.to(device="cpu") for arg in args]
            device = "mps"
        else:
            device = None

        dtypes = [arg.dtype for arg in args]
        dtype = reduce(torch.promote_types, dtypes)

        args = [arg.to(torch.promote_types(arg.dtype, min_dtype)) for arg in args]
        outs = f(*args, **kwargs)

        if torch.is_tensor(outs):
            return outs.to(dtype=dtype, device=device)
        else:
            return tuple(out.to(dtype=dtype, device=device) for out in outs)

    return g



r"""Elucidated latent diffusion model (ELDM or EDM2) plugin.

This plugin depends on the `torch_utils` and `training` modules in the `NVlabs/edm2
<https://github.com/NVlabs/edm2>`_ repository. To use it, clone the repository to your
machine

.. code-block:: console

    git clone https://github.com/NVlabs/edm2

and add it to your Python path before importing the plugin.

.. code-block:: python

    import sys; sys.path.append("path/to/edm2")
    ...
    from azula.plugins import eldm

You may also need to install additional dependencies in your environment, including
:mod:`diffusers` and :mod:`accelerate`.

.. code-block:: console

    pip install diffusers accelerate

References:
    | Analyzing and Improving the Training Dynamics of Diffusion Models (Karras et al., 2024)
    | https://arxiv.org/abs/2312.02696
"""

__all__ = [
    "AutoEncoder",
    "ElucidatedLatentDenoiser",
    "load_model",
]

import pickle
import torch
import torch.nn as nn

from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.hub import download
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import Schedule

from ..edm import ElucidatedSchedule
from ..utils import load_cards, patch_diffusers, patch_mmap_safetensors


class AutoEncoder(nn.Module):
    r"""Creates an auto-encoder wrapper."""

    def __init__(
        self,
        vae: nn.Module,  # AutoencoderKL
        shift: Tensor,
        scale: Tensor,
    ) -> None:
        super().__init__()

        self.vae = vae

        self.register_buffer("shift", torch.as_tensor(shift))
        self.register_buffer("scale", torch.as_tensor(scale))

    def encode(self, x: Tensor) -> Tensor:
        r"""Encodes images to latents.

        Arguments:
            x: A batch of images :math:`x`, with shape :math:`(B, 3, 512, 512)`.
                Pixel values are expected to range between 0 and 1.

        Returns:
            A batch of latents :math:`z \sim q(Z \mid x)`, with shape :math:`(B, 4, 64, 64)`.
        """

        dtype = get_module_dtype(self.vae)

        q_z_x = self.vae.encode(x.to(dtype)).latent_dist
        z = q_z_x.mean + q_z_x.std * torch.randn_like(q_z_x.mean)
        z = z * self.scale + self.shift

        return z.to(x)

    def decode(self, z: Tensor) -> Tensor:
        r"""Decodes latents to images.

        Arguments:
            z: A batch of latents :math:`z`, with shape :math:`(B, 4, 64, 64)`.

        Returns:
            A batch of images :math:`x = D(z)`, with shape :math:`(B, 3, 512, 512)`.
        """

        dtype = get_module_dtype(self.vae)

        z = (z - self.shift) / self.scale
        x = self.vae.decode(z.to(dtype)).sample

        return x.to(z)


class ElucidatedLatentDenoiser(Denoiser):
    r"""Creates an elucidated latent denoiser.

    Arguments:
        backbone: A noise conditional network.
        schedule: A noise schedule. If :py:`None`, use
            :class:`azula.plugins.edm.ElucidatedSchedule` instead.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = ElucidatedSchedule()
        else:
            self.schedule = schedule

    def forward(
        self,
        z_t: Tensor,
        t: Tensor,
        label: Tensor | None = None,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            z_t: A noisy tensor :math:`z_t`, with shape :math:`(B, 4, 64, 64)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            label: The class label :math:`c` as a one-hot vector, with shape :math:`(*, 1000)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(Z - \mu_\phi(z_t \mid c))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < z_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = 1 / alpha_t
        c_time = (sigma_t / alpha_t).reshape_as(t)

        dtype = get_module_dtype(self.backbone)

        mean = self.backbone(
            (c_in * z_t).to(dtype),
            c_time.to(dtype),
            class_labels=label.to(dtype),
            **kwargs,
        ).to(z_t)

        return DiracPosterior(mean=mean)


def load_model(name: str) -> tuple[Denoiser, AutoEncoder]:
    r"""Loads a pre-trained ELDM (or EDM2) latent denoiser.

    Arguments:
        name: The pre-trained model name.

    Returns:
        A pre-trained latent denoiser and the corresponding auto-encoder.
    """

    from diffusers import AutoencoderKL

    card = load_cards(__name__)[name]

    with open(download(card.url, hash_prefix=card.hash), "rb") as f:
        content = pickle.load(f)

    denoiser = content["ema"]
    denoiser = ElucidatedLatentDenoiser(
        backbone=denoiser,
    )

    with skip_init(), patch_diffusers():
        vae = AutoencoderKL.from_pretrained("stabilityai/sd-vae-ft-mse")

    patch_mmap_safetensors(vae)

    autoencoder = content["encoder"]
    autoencoder = AutoEncoder(
        vae=vae,
        shift=autoencoder.bias.reshape(-1, 1, 1),
        scale=autoencoder.scale.reshape(-1, 1, 1),
    )

    return denoiser.eval(), autoencoder.eval()



r"""Tweedie Moment Projected Diffusion (TMPD) internals.

References:
    | Tweedie Moment Projected Diffusions For Inverse Problems (Boys et al., 2023)
    | https://arxiv.org/abs/2310.06721
"""

__all__ = [
    "TMPDenoiser",
]

import torch

from collections.abc import Callable
from torch import Tensor

from ..denoise import Denoiser, DiracPosterior
from ..noise import Schedule


class TMPDenoiser(Denoiser):
    r"""Creates a TMPD denoiser module.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y \sim \mathcal{N}(A x, \Sigma_y)`.
        A: The forward operator :math:`x \mapsto A x`.
        var_y: The noise variance :math:`\Sigma_y`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        A: Callable[[Tensor], Tensor],
        var_y: float | Tensor,
    ) -> None:
        super().__init__()

        self.denoiser = denoiser

        self.y = y
        self.A = A
        self.var_y = var_y

    @property
    def schedule(self) -> Schedule:
        return self.denoiser.schedule

    @torch.no_grad()
    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        alpha_t, sigma_t = self.schedule(t)
        gamma_t = sigma_t**2 / alpha_t

        with torch.enable_grad():
            x_t = x_t.detach().requires_grad_()
            q = self.denoiser(x_t, t, **kwargs)

            x_hat = q.mean
            y_hat = self.A(x_hat)

        def At(v: Tensor) -> Tensor:
            return torch.autograd.grad(y_hat, x_hat, v, retain_graph=True)[0]

        def cov_x(v: Tensor) -> Tensor:
            return gamma_t * torch.autograd.grad(x_hat, x_t, v, retain_graph=True)[0]

        var_Ax = self.A(cov_x(At(torch.ones_like(y_hat))))

        grad = (self.y - y_hat) / (self.var_y + var_Ax)
        grad = gamma_t * torch.autograd.grad(y_hat, x_t, grad)[0]

        return DiracPosterior(mean=x_hat + grad)



r"""Ablated diffusion model (ADM) plugin.

.. code-block:: python

    from azula.plugins import adm

References:
    | Diffusion Models Beat GANs on Image Synthesis (Dhariwal et al., 2021)
    | https://arxiv.org/abs/2105.05233
"""

__all__ = [
    "AblatedDenoiser",
    "load_model",
]

import torch
import torch.nn as nn

from collections.abc import Sequence
from torch import Tensor

from azula.denoise import Denoiser, GaussianPosterior
from azula.hub import download
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import Schedule, VPSchedule

from ._src import unet
from ..utils import load_cards


class AblatedDenoiser(Denoiser):
    r"""Creates an ablated denoiser.

    Arguments:
        backbone: A time conditional network.
        schedule: A noise schedule. If :py:`None`, use :class:`azula.noise.VPSchedule`
            instead.
        clip_mean: Whether the mean :math:`\mu_\phi(x_t)` is clipped to :math:`[-1, 1]`
            or not during evaluation.
        learn_var: Whether the variance :math:`\sigma^2_\phi(x_t)` is learned or not.
            For pre-trained models, the learned variance is indicative, but inexact.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
        clip_mean: bool = False,
        learn_var: bool = False,
        discrete_schedule: str = "linear",
        discrete_steps: int = 1000,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = VPSchedule(alpha_min=1e-2, sigma_min=1e-2)
        else:
            self.schedule = schedule

        self.clip_mean = clip_mean
        self.learn_var = learn_var

        if discrete_schedule == "linear":
            beta = torch.linspace(
                0.1 / discrete_steps,
                20.0 / discrete_steps,
                discrete_steps,
                dtype=torch.float64,
            )
        elif discrete_schedule == "cosine":
            t = torch.linspace(0, 1, discrete_steps + 1, dtype=torch.float64)
            alpha_bar = torch.cos((t + 0.008) / 1.008 * torch.pi / 2) ** 2
            beta = 1 - alpha_bar[1:] / alpha_bar[:-1]
            beta = torch.clip(beta, max=0.999)
        else:
            raise ValueError(f"Unknown discrete schedule '{discrete_schedule}'.")

        alpha_bar = torch.cumprod(1 - beta, dim=0)
        sigmas = torch.sqrt(1 - alpha_bar)

        self.register_buffer("sigmas", sigmas.to(torch.get_default_dtype()))

    def forward(
        self,
        x_t: Tensor,
        t: Tensor,
        label: Tensor | None = None,
        **kwargs,
    ) -> GaussianPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, 3, H, W)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            label: The class label :math:`c` as an integer, with shape :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Gaussian :math:`\mathcal{N}(X \mid \mu_\phi(x_t \mid c), \sigma^2_\phi(x_t \mid c))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_out = -sigma_t / alpha_t
        c_skip = 1 / alpha_t
        c_time = sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_time = torch.searchsorted(self.sigmas, c_time.flatten())
        c_var = sigma_t**2 / (alpha_t**2 + sigma_t**2)

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            (c_in * x_t).to(dtype),
            c_time,
            y=label,
            **kwargs,
        ).to(x_t)

        if self.learn_var:
            output, log_var = torch.chunk(output, 2, dim=1)
            mean = c_skip * x_t + c_out * output
            var = c_var * torch.exp(log_var)
        else:
            mean = c_skip * x_t + c_out * output
            var = c_var

        if not self.training and self.clip_mean:
            mean = torch.clip(mean, min=-1.0, max=1.0)

        return GaussianPosterior(mean=mean, var=var)


def load_model(name: str, **kwargs) -> Denoiser:
    r"""Loads a pre-trained ADM denoiser.

    Arguments:
        name: The pre-trained model name.
        kwargs: Keyword arguments passed to :func:`torch.load`.

    Returns:
        A pre-trained denoiser.
    """

    kwargs.setdefault("map_location", "cpu")
    kwargs.setdefault("weights_only", True)

    card = load_cards(__name__)[name]
    state = torch.load(download(card.url, hash_prefix=card.hash), **kwargs)

    with skip_init():
        denoiser = make_model(**card.config)

    denoiser.backbone.load_state_dict(state)

    return denoiser.eval()


def make_model(
    # Denoiser
    clip_mean: bool = True,
    learn_var: bool = True,
    # Discrete schedule
    discrete_schedule: str = "linear",
    discrete_steps: int = 1000,
    # Data
    image_channels: int = 3,
    image_size: int = 64,
    # Backbone
    attention_resolutions: Sequence[int] = (32, 16, 8),
    channel_mult: Sequence[int] = (1, 2, 3, 4),
    num_channels: int = 128,
    num_classes: int | None = None,
    **kwargs,
) -> Denoiser:
    r"""Initializes an ADM denoiser."""

    attention_resolutions = {image_size // r for r in attention_resolutions}

    backbone = unet.UNetModel(
        image_size=image_size,
        in_channels=image_channels,
        out_channels=2 * image_channels if learn_var else image_channels,
        model_channels=num_channels,
        channel_mult=channel_mult,
        num_classes=num_classes,
        attention_resolutions=attention_resolutions,
        **kwargs,
    )

    return AblatedDenoiser(
        backbone,
        clip_mean=clip_mean,
        learn_var=learn_var,
        discrete_schedule=discrete_schedule,
        discrete_steps=discrete_steps,
    )



r"""Tests for the azula.noise module."""

import pytest
import torch

from collections.abc import Sequence

from azula.noise import CosineSchedule, DecaySchedule, RectifiedSchedule, VESchedule, VPSchedule


@pytest.mark.parametrize("batch", [(), (64,)])
def test_schedules(batch: Sequence[int]) -> None:
    Ss = [
        VPSchedule,
        VESchedule,
        CosineSchedule,
        RectifiedSchedule,
        DecaySchedule,
    ]

    for S in Ss:
        schedule = S()

        # Scales
        t = torch.rand(batch)
        alpha_t, sigma_t = schedule(t)

        assert alpha_t.shape == batch, S
        assert sigma_t.shape == batch, S

        assert torch.all(alpha_t > 0), S
        assert torch.all(sigma_t > 0), S

        # Monotonicity
        s = torch.rand_like(t) * t
        alpha_s, sigma_s = schedule(s)

        assert torch.all(alpha_s / sigma_s >= alpha_t / sigma_t), S

        # Start time
        alpha_0, _ = schedule(torch.zeros(()))

        assert torch.all(alpha_0 == 1), S



r"""Tests for the azula.denoise module."""

import pytest
import torch

from collections.abc import Sequence
from torch import Tensor
from torch.distributions import Normal

from azula.denoise import (
    GaussianDenoiser,
    GaussianPosterior,
    KarrasDenoiser,
    Posterior,
    SimpleDenoiser,
)
from azula.linalg.covariance import DPLRCovariance, KroneckerCovariance
from azula.nn.layers import SineEncoding
from azula.noise import RectifiedSchedule, Schedule, VPSchedule


class Dummy(torch.nn.Module):
    def __init__(self, features: int = 5, with_label: bool = False) -> None:
        super().__init__()

        self.with_label = with_label

        self.l1 = torch.nn.Linear(features, 64)
        self.l2 = torch.nn.Linear(64, features)
        self.relu = torch.nn.ReLU()

        self.time_encoding = SineEncoding(64)

    def forward(self, x_t: Tensor, t: Tensor, label: str | None = None) -> Tensor:
        y = self.l1(x_t)
        y = y + self.time_encoding(t)
        y = self.relu(y)
        y = self.l2(y)

        if self.with_label:
            assert isinstance(label, str)
        else:
            assert label is None

        return y


@pytest.mark.parametrize("isotropic", [False, True])
@pytest.mark.parametrize("batch", [(), (64,)])
@pytest.mark.parametrize("channels", [5])
def test_GaussianPosterior(isotropic: bool, batch: Sequence[int], channels: int) -> None:
    mean = torch.randn(*batch, channels)

    if isotropic:
        std = torch.rand(*batch, 1) + 1e-3
    else:
        std = torch.rand(*batch, channels) + 1e-3

    x = mean + std * torch.randn_like(mean)

    log_q = GaussianPosterior(mean, std**2).log_prob(x)
    log_p = Normal(mean, std).log_prob(x)

    assert log_q.shape == (*batch, channels)
    assert torch.allclose(log_q, log_p, atol=1e-6)


@pytest.mark.parametrize("cov", ["dplr", "precond"])
@pytest.mark.parametrize("batch", [(), (64,)])
@pytest.mark.parametrize("channels", [5])
def test_GaussianDenoiser(cov: str, batch: Sequence[int], channels: int) -> None:
    data = torch.randn(256, channels)
    mean = torch.mean(data, dim=0)

    if cov == "dplr":
        cov = DPLRCovariance.from_data(data, rank=3)
    elif cov == "precond":
        cov = KroneckerCovariance.from_data(data, rank=0)

    denoiser = GaussianDenoiser(mean, cov, schedule=VPSchedule())

    # Forward
    x = torch.randn(*batch, channels, requires_grad=True)
    t = torch.rand(())

    q = denoiser(x, t)

    assert isinstance(q, Posterior)
    assert q.mean.shape == x.shape


class ReSchedule(Schedule):
    def __init__(self, schedule: Schedule) -> None:
        self.schedule = schedule

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        alpha, sigma = self.schedule(t)
        return torch.ones_like(alpha), sigma / alpha


@pytest.mark.parametrize("denoiser_cls", [SimpleDenoiser, KarrasDenoiser])
@pytest.mark.parametrize("schedule_cls", [VPSchedule, RectifiedSchedule])
@pytest.mark.parametrize("with_label", [False, True])
@pytest.mark.parametrize("batch", [(), (64,)])
@pytest.mark.parametrize("channels", [5])
def test_denoisers(
    denoiser_cls: type,
    schedule_cls: type,
    with_label: bool,
    batch: Sequence[int],
    channels: int,
) -> None:
    denoiser = denoiser_cls(
        backbone=Dummy(channels, with_label),
        schedule=schedule_cls(),
    )

    # Forward
    x = torch.randn(*batch, channels)
    t = torch.rand(batch)

    alpha_t, sigma_t = denoiser.schedule(t)
    alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

    x_t = torch.normal(alpha_t * x, sigma_t)

    if with_label:
        q = denoiser(x_t, t, label="cat")
    else:
        q = denoiser(x_t, t)

    assert isinstance(q, Posterior)
    assert q.mean.shape == x.shape

    ## Reschedule to VE
    denoiser.schedule = ReSchedule(denoiser.schedule)

    if with_label:
        q_ve = denoiser(x_t / alpha_t, t, label="cat")
    else:
        q_ve = denoiser(x_t / alpha_t, t)

    assert torch.allclose(q.mean, q_ve.mean, atol=1e-6)

    # Loss
    if with_label:
        loss = denoiser.loss(x, t, label="cat")
    else:
        loss = denoiser.loss(x, t)

    assert loss.shape == ()
    assert loss.requires_grad

    loss.mean().backward()

    for p in denoiser.parameters():
        assert p.grad is not None
        assert torch.all(torch.isfinite(p.grad))



import math
import torch

from torch import nn

from . import utils


class ResidualBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.skip = skip if skip else nn.Identity()
        self.main = nn.Sequential(*main)

    def forward(self, input):
        return self.main(input) + self.skip(input)


class ResConvBlock(ResidualBlock):
    def __init__(self, c_in, c_mid, c_out, is_last=False) -> None:
        skip = None if c_in == c_out else nn.Conv2d(c_in, c_out, 1, bias=False)
        super().__init__(
            [
                nn.Conv2d(c_in, c_mid, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(c_mid, c_out, 3, padding=1),
                nn.ReLU(inplace=True) if not is_last else nn.Identity(),
            ],
            skip,
        )


class SelfAttention2d(nn.Module):
    def __init__(self, c_in, n_head=1) -> None:
        super().__init__()
        assert c_in % n_head == 0
        self.n_head = n_head
        self.qkv_proj = nn.Conv2d(c_in, c_in * 3, 1)
        self.out_proj = nn.Conv2d(c_in, c_in, 1)

    def forward(self, input):
        n, c, h, w = input.shape
        qkv = self.qkv_proj(input)
        qkv = qkv.view([n, self.n_head * 3, c // self.n_head, h * w]).transpose(2, 3)
        q, k, v = qkv.chunk(3, dim=1)
        scale = k.shape[3] ** -0.25
        att = ((q * scale) @ (k.transpose(2, 3) * scale)).softmax(3)
        y = (att @ v).transpose(2, 3).contiguous().view([n, c, h, w])
        return input + self.out_proj(y)


class SkipBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return torch.cat([self.skip(input), self.main(input)], dim=1)


class FourierFeatures(nn.Module):
    def __init__(self, in_features, out_features, std=1.0) -> None:
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(torch.randn([out_features // 2, in_features]) * std)

    def forward(self, input):
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)


def expand_to_planes(input, shape):
    return input[..., None, None].repeat([1, 1, shape[2], shape[3]])


class WikiArt256Model(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.shape = (3, 256, 256)
        self.min_t = utils.get_ddpm_schedule(torch.tensor(0.0)).item()
        self.max_t = utils.get_ddpm_schedule(torch.tensor(1.0)).item()

        c = 128  # The base channel count
        cs = [c // 2, c, c * 2, c * 2, c * 4, c * 4, c * 8]

        self.timestep_embed = FourierFeatures(1, 16, std=0.2)
        self.down = nn.AvgPool2d(2)
        self.up = nn.Upsample(scale_factor=2, mode="nearest")

        self.net = nn.Sequential(  # 256x256
            ResConvBlock(3 + 16, cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            SkipBlock([
                self.down,  # 256x256 -> 128x128
                ResConvBlock(cs[0], cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[1]),
                SkipBlock([
                    self.down,  # 128x128 -> 64x64
                    ResConvBlock(cs[1], cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[2]),
                    SkipBlock([
                        self.down,  # 64x64 -> 32x32
                        ResConvBlock(cs[2], cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[3]),
                        SkipBlock([
                            self.down,  # 32x32 -> 16x16
                            ResConvBlock(cs[3], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            SkipBlock([
                                self.down,  # 16x16 -> 8x8
                                ResConvBlock(cs[4], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                SkipBlock([
                                    self.down,  # 8x8 -> 4x4
                                    ResConvBlock(cs[5], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[6]),
                                    SelfAttention2d(cs[6], cs[6] // 128),
                                    ResConvBlock(cs[6], cs[6], cs[5]),
                                    SelfAttention2d(cs[5], cs[5] // 128),
                                    self.up,  # 4x4 -> 8x8
                                ]),
                                ResConvBlock(cs[5] * 2, cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[5]),
                                SelfAttention2d(cs[5], cs[5] // 128),
                                ResConvBlock(cs[5], cs[5], cs[4]),
                                SelfAttention2d(cs[4], cs[4] // 128),
                                self.up,  # 8x8 -> 16x16
                            ]),
                            ResConvBlock(cs[4] * 2, cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[4]),
                            SelfAttention2d(cs[4], cs[4] // 128),
                            ResConvBlock(cs[4], cs[4], cs[3]),
                            SelfAttention2d(cs[3], cs[3] // 128),
                            self.up,  # 16x16 -> 32x32
                        ]),
                        ResConvBlock(cs[3] * 2, cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[3]),
                        ResConvBlock(cs[3], cs[3], cs[2]),
                        self.up,  # 32x32 -> 64x64
                    ]),
                    ResConvBlock(cs[2] * 2, cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[2]),
                    ResConvBlock(cs[2], cs[2], cs[1]),
                    self.up,  # 64x64 -> 128x128
                ]),
                ResConvBlock(cs[1] * 2, cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[1]),
                ResConvBlock(cs[1], cs[1], cs[0]),
                self.up,  # 128x128 -> 256x256
            ]),
            ResConvBlock(cs[0] * 2, cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], cs[0]),
            ResConvBlock(cs[0], cs[0], 3, is_last=True),
        )

    def forward(self, input, t):
        log_snr = utils.alpha_sigma_to_log_snr(*utils.t_to_alpha_sigma(t))
        timestep_embed = expand_to_planes(self.timestep_embed(log_snr[:, None]), input.shape)
        return self.net(torch.cat([input, timestep_embed], dim=1))



r"""Vision Transformer (ViT) building blocks.

References:
    | An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale (Dosovitskiy et al., 2021)
    | https://arxiv.org/abs/2010.11929
"""

__all__ = [
    "ViT",
]

import math
import torch

from collections.abc import Sequence
from torch import Tensor

from .dit import DiT
from .layers import Patchify, Unpatchify


class ViT(DiT):
    r"""Creates a modulated ViT-like module.

    Arguments:
        in_channels: The number of input channels :math:`C_i`.
        out_channels: The number of output channels :math:`C_o`.
        cond_channels: The number of condition channels :math:`C_c`.
        mod_features: The number of modulating features :math:`D`.
        hid_channels: The numbers of hidden token channels.
        hid_blocks: The number of hidden transformer blocks.
        spatial: The number of spatial dimensions :math:`N`.
        patch_size: The patch size or shape.
        unpatch_size: The unpatch size or shape.
        kwargs: Keyword arguments passed to :class:`ViTBlock`.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        cond_channels: int = 0,
        mod_features: int = 0,
        hid_channels: int = 1024,
        hid_blocks: int = 3,
        spatial: int = 2,
        patch_size: int | Sequence[int] = 1,
        unpatch_size: int | Sequence[int] | None = None,
        **kwargs,
    ) -> None:
        if isinstance(patch_size, int):
            patch_size = [patch_size] * spatial

        if unpatch_size is None:
            unpatch_size = patch_size
        elif isinstance(unpatch_size, int):
            unpatch_size = [unpatch_size] * spatial

        assert len(patch_size) == len(unpatch_size) == spatial

        super().__init__(
            in_channels=math.prod(patch_size) * in_channels,
            out_channels=math.prod(unpatch_size) * out_channels,
            cond_channels=math.prod(patch_size) * cond_channels,
            mod_features=mod_features,
            pos_channels=spatial,
            hid_channels=hid_channels,
            hid_blocks=hid_blocks,
            **kwargs,
        )

        self.patch = Patchify(patch_size, channel_last=True)
        self.unpatch = Unpatchify(unpatch_size, channel_last=True)
        self.spatial = spatial

    def forward(
        self,
        x: Tensor,
        mod: Tensor | None = None,
        cond: Tensor | None = None,
    ) -> Tensor:
        r"""
        Arguments:
            x: The input tensor, with shape :math:`(B, C_i, L_1, ..., L_N)`.
            mod: The modulation vector, with shape :math:`(D)` or :math:`(B, D)`.
            cond: The condition tensor, with :math:`(B, C_c, L_1, ..., L_N)`.

        Returns:
            The output tensor, with shape :math:`(B, C_o, L_1, ..., L_N)`.
        """

        x = self.patch(x)

        if cond is not None:
            cond = self.patch(cond)

        shape = x.shape[1:-1]

        pos = (torch.arange(size, dtype=x.dtype, device=x.device) for size in shape)
        pos = torch.cartesian_prod(*pos)
        pos = torch.reshape(pos, shape=(-1, len(shape)))

        x = x.flatten(1, -2)
        y = super().forward(x, mod, pos=pos, cond=cond)
        y = y.unflatten(-2, shape)
        y = self.unpatch(y)

        return y



# Adapted from
# https://github.com/crowsonkb/v-diffusion-pytorch/blob/ffabbb1a897541fa2a3d034f397c224489d97b39/diffusion/utils.py

import io
import math
import requests
import torch

from torchvision.transforms import functional as TF


def fetch(url_or_path):
    """Fetches a file from an HTTP or HTTPS url, or opens the local file."""
    if str(url_or_path).startswith("http://") or str(url_or_path).startswith("https://"):
        r = requests.get(url_or_path)
        r.raise_for_status()
        fd = io.BytesIO()
        fd.write(r.content)
        fd.seek(0)
        return fd
    return open(url_or_path, "rb")


def from_pil_image(x):
    """Converts from a PIL image to a tensor."""
    x = TF.to_tensor(x)
    if x.ndim == 2:
        x = x[..., None]
    return x * 2 - 1


def to_pil_image(x):
    """Converts from a tensor to a PIL image."""
    if x.ndim == 4:
        assert x.shape[0] == 1
        x = x[0]
    if x.shape[0] == 1:
        x = x[0]
    return TF.to_pil_image((x.clamp(-1, 1) + 1) / 2)


def log_snr_to_alpha_sigma(log_snr):
    """Returns the scaling factors for the clean image and for the noise, given
    the log SNR for a timestep."""
    return log_snr.sigmoid().sqrt(), log_snr.neg().sigmoid().sqrt()


def alpha_sigma_to_log_snr(alpha, sigma):
    """Returns a log snr, given the scaling factors for the clean image and for
    the noise."""
    return torch.log(alpha**2 / sigma**2)


def t_to_alpha_sigma(t):
    """Returns the scaling factors for the clean image and for the noise, given
    a timestep."""
    return torch.cos(t * math.pi / 2), torch.sin(t * math.pi / 2)


def alpha_sigma_to_t(alpha, sigma):
    """Returns a timestep, given the scaling factors for the clean image and for
    the noise."""
    return torch.atan2(sigma, alpha) / math.pi * 2


def get_ddpm_schedule(ddpm_t):
    """Returns timesteps for the noise schedule from the DDPM paper."""
    log_snr = -torch.special.expm1(1e-4 + 10 * ddpm_t**2).log()
    alpha, sigma = log_snr_to_alpha_sigma(log_snr)
    return alpha_sigma_to_t(alpha, sigma)


def get_spliced_ddpm_cosine_schedule(t):
    """Returns timesteps for a spliced DDPM/cosine noise schedule."""
    ddpm_crossover = 0.48536712
    cosine_crossover = 0.80074257
    big_t = t * (1 + cosine_crossover - ddpm_crossover)
    ddpm_part = get_ddpm_schedule(big_t + ddpm_crossover - cosine_crossover)
    return torch.where(big_t < cosine_crossover, big_t, ddpm_part)


def get_log_schedule(t, min_log_snr=-10, max_log_snr=10):
    """Returns timesteps for a logarithmically spaced schedule."""
    log_snr = t * (min_log_snr - max_log_snr) + max_log_snr
    alpha, sigma = log_snr_to_alpha_sigma(log_snr)
    return alpha_sigma_to_t(alpha, sigma)



r"""Miscelaneous plugin helpers."""

__all__ = [
    "load_cards",
]

import os
import sys
import torch
import yaml

from collections.abc import Iterator
from contextlib import contextmanager
from types import ModuleType, SimpleNamespace


def as_torch_dtype(name: str | None = None) -> torch.dtype:
    if name is None:
        return None

    dtype = getattr(torch, name, None)

    if isinstance(dtype, torch.dtype):
        return dtype
    else:
        raise ValueError(f"Unknown data type '{name}'.")


def load_cards(plugin: ModuleType) -> dict[str, SimpleNamespace]:
    r"""Returns the name-card mapping of pre-trained models available in a plugin.

    Arguments:
        plugin: The plugin module.

    Example:
        >>> cards = load_cards(azula.plugins.adm)
        >>> list(cards.keys())
        ['imagenet_64x64_cond',
         'imagenet_128x128_cond',
         'imagenet_256x256',
         'imagenet_256x256_cond',
         'imagenet_512x512_cond',
         'ffhq_256x256']
    """

    if isinstance(plugin, str):
        plugin = sys.modules[plugin]

    file = os.path.join(os.path.dirname(plugin.__file__), "cards.yaml")

    assert os.path.exists(file), f"{plugin} is not a plugin"

    with open(file) as f:
        cards = yaml.safe_load(f)

    for card in cards.values():
        if "dtype_map" in card:
            card["dtype_map"] = {k: as_torch_dtype(v) for k, v in card["dtype_map"].items()}

    return {name: SimpleNamespace(**card) for name, card in cards.items()}


@contextmanager
def patch_diffusers() -> Iterator[None]:
    from unittest.mock import patch

    with (
        patch("transformers.modeling_utils.logger.warning"),
        patch("transformers.modeling_utils.logger.warning_once"),
    ):
        yield


def patch_mmap_safetensors(*modules: torch.nn.Module) -> None:
    for m in modules:
        for p in m.parameters():
            p.data.sum()

        for b in m.buffers():
            b.data.sum()



r"""Elucidated diffusion model (EDM) plugin.

This plugin depends on the `torch_utils` and `training` modules in the `NVlabs/edm
<https://github.com/NVlabs/edm>`_ repository. To use it, clone the repository to your
machine

.. code-block:: console

    git clone https://github.com/NVlabs/edm

and add it to your Python path before importing the plugin.

.. code-block:: python

    import sys; sys.path.append("path/to/edm")
    ...
    from azula.plugins import edm

References:
    | Elucidating the Design Space of Diffusion-Based Generative Models (Karras et al., 2022)
    | https://arxiv.org/abs/2206.00364
"""

__all__ = [
    "ElucidatedSchedule",
    "ElucidatedDenoiser",
    "load_model",
]

import pickle
import torch
import torch.nn as nn

from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.hub import download
from azula.nn.utils import get_module_dtype
from azula.noise import Schedule

from ..utils import load_cards


class ElucidatedSchedule(Schedule):
    r"""Creates an elucidated noise schedule.

    .. math::
        \alpha_t & = 1 \\
        \sigma_t & = \left( (1 - t) \, {\sigma_\min}^\frac{1}{\rho}
            + t \, {\sigma_\max}^\frac{1}{\rho} \right)^\rho

    Arguments:
        sigma_min: The initial noise scale :math:`\sigma_\min \in \mathbb{R}_+`.
        sigma_max: The final noise scale :math:`\sigma_\max \in \mathbb{R}_+`.
        rho: A hyper-parameter :math:`\rho \in \mathbb{R}_+`.
    """

    def __init__(
        self, sigma_min: float = 0.002, sigma_max: float = 80.0, rho: float = 7.0
    ) -> None:
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.rho = rho

    def alpha(self, t: Tensor) -> Tensor:
        return torch.ones_like(t)

    def sigma(self, t: Tensor) -> Tensor:
        lower = self.sigma_min ** (1 / self.rho)
        upper = self.sigma_max ** (1 / self.rho)
        return torch.pow((1 - t) * lower + t * upper, self.rho)

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)


class ElucidatedDenoiser(Denoiser):
    r"""Creates an elucidated denoiser.

    Arguments:
        backbone: A noise conditional network.
        schedule: A noise schedule. If :py:`None`, use :class:`ElucidatedSchedule` instead.
    """

    def __init__(self, backbone: nn.Module, schedule: Schedule | None = None) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = ElucidatedSchedule()
        else:
            self.schedule = schedule

    def forward(
        self,
        x_t: Tensor,
        t: Tensor,
        label: Tensor | None = None,
        **kwargs,
    ) -> DiracPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, 3, H, W)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            label: The class label :math:`c` as a one-hot vector, with shape :math:`(*, 1000)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \mu_\phi(x_t \mid c))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = 1 / alpha_t
        c_time = (sigma_t / alpha_t).reshape_as(t)

        dtype = get_module_dtype(self.backbone)

        mean = self.backbone(
            (c_in * x_t).to(dtype),
            c_time.to(dtype),
            class_labels=label.to(dtype),
            **kwargs,
        ).to(x_t)

        return DiracPosterior(mean=mean)


def load_model(name: str) -> Denoiser:
    r"""Loads a pre-trained EDM denoiser.

    Arguments:
        name: The pre-trained model name.

    Returns:
        A pre-trained denoiser.
    """

    card = load_cards(__name__)[name]

    with open(download(card.url, hash_prefix=card.hash), "rb") as f:
        content = pickle.load(f)

    denoiser = content["ema"]
    denoiser = ElucidatedDenoiser(
        backbone=denoiser,
    )

    return denoiser.eval()



r"""Utilities for downloading models."""

__all__ = [
    "get_hub_dir",
    "set_hub_dir",
    "download",
]

import gdown
import hashlib
import os
import re
import shutil
import sys
import tarfile
import tempfile
import torch
import zipfile

AZULA_HUB: str = os.path.expanduser("~/.cache/azula/hub")


def get_hub_dir() -> str:
    r"""Returns the cache directory used for storing models & weights."""

    return AZULA_HUB


def set_hub_dir(cache_dir: str) -> None:
    r"""Sets the cache directory used for storing models & weights."""

    global AZULA_HUB

    cache_dir = os.path.expanduser(cache_dir)
    cache_dir = os.path.abspath(cache_dir)

    AZULA_HUB = cache_dir


def download(
    url: str,
    filename: str | None = None,
    hash_prefix: str | None = None,
    extract: bool = False,
    quiet: bool = False,
) -> str:
    r"""Downloads data at a given URL to a local file.

    Arguments:
        url: A URL. Google Drive URLs are supported.
        filename: A local file name. If :py:`None`, use the sanitized URL instead.
            If a file with the same name exists, the download is skipped.
        hash_prefix: The expected hash prefix of the file, formatted as `"alg:prefix"`.
        extract: Whether to extract archives or not.
        quiet: Whether to keep it quiet in the terminal or not.

    Returns:
        The local file name. If :py:`extract=True`, the local directory name containing
        the extracted files is returned instead.
    """

    if filename is None:
        filename = re.sub(r"[^a-zA-Z0-9_]+", ".", url)
        filename = os.path.join(get_hub_dir(), filename)
    else:
        filename = os.path.expanduser(filename)
        filename = os.path.abspath(filename)

    os.makedirs(os.path.dirname(filename), exist_ok=True)

    if os.path.exists(filename):
        if not quiet:
            print(f"Loading from {filename}", file=sys.stderr)
    else:
        if not quiet:
            print(f"Downloading {url} to {filename}", file=sys.stderr)

        if "drive.google" in url:
            gdown.download(url, filename, quiet=quiet)
        else:
            torch.hub.download_url_to_file(url, filename, progress=not quiet)

    if hash_prefix is not None:
        alg, prefix = hash_prefix.split(":")
        digest = hashlib.new(alg)

        with open(filename, "rb") as f:  # adapted from hashlib.file_digest
            buffer = bytearray(2**20)  # reusable 1MB buffer
            view = memoryview(buffer)
            while True:
                size = f.readinto(buffer)
                if size == 0:  # end of file
                    break
                digest.update(view[:size])

        hex_hash = digest.hexdigest()

        assert hex_hash.startswith(prefix), (
            f"The hash of the downloaded file ({alg}:{hex_hash}) does not match "
            f"the expected hash prefix ({alg}:{prefix})."
        )

    if extract:
        xd = f"{filename}+x"

        if os.path.exists(xd):
            return xd
        elif not quiet:
            print(f"Extracting to {xd}", file=sys.stderr)

        with tempfile.TemporaryDirectory() as td:
            if tarfile.is_tarfile(filename):
                with tarfile.TarFile(filename, "r") as f:
                    f.extractall(td)
            elif zipfile.is_zipfile(filename):
                with zipfile.ZipFile(filename, "r") as f:
                    f.extractall(td)
            else:
                raise ValueError("Unknown archive format.")

            shutil.move(td, xd)

        return xd

    return filename



# Adapted from
# https://github.com/openai/guided-diffusion/blob/aca3097a89fab9c9962b22224d5f48950eb57e29/guided_diffusion/nn.py

import math
import torch as th
import torch.nn as nn


def conv_nd(dims, *args, **kwargs):
    """
    Create a 1D, 2D, or 3D convolution module.
    """
    if dims == 1:
        return nn.Conv1d(*args, **kwargs)
    elif dims == 2:
        return nn.Conv2d(*args, **kwargs)
    elif dims == 3:
        return nn.Conv3d(*args, **kwargs)
    raise ValueError(f"unsupported dimensions: {dims}")


def linear(*args, **kwargs):
    """
    Create a linear module.
    """
    return nn.Linear(*args, **kwargs)


def avg_pool_nd(dims, *args, **kwargs):
    """
    Create a 1D, 2D, or 3D average pooling module.
    """
    if dims == 1:
        return nn.AvgPool1d(*args, **kwargs)
    elif dims == 2:
        return nn.AvgPool2d(*args, **kwargs)
    elif dims == 3:
        return nn.AvgPool3d(*args, **kwargs)
    raise ValueError(f"unsupported dimensions: {dims}")


def update_ema(target_params, source_params, rate=0.99) -> None:
    """
    Update target parameters to be closer to those of source parameters using
    an exponential moving average.

    :param target_params: the target parameter sequence.
    :param source_params: the source parameter sequence.
    :param rate: the EMA rate (closer to 1 means slower).
    """
    for targ, src in zip(target_params, source_params, strict=True):
        targ.detach().mul_(rate).add_(src, alpha=1 - rate)


def zero_module(module):
    """
    Zero out the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


def scale_module(module, scale):
    """
    Scale the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().mul_(scale)
    return module


def mean_flat(tensor):
    """
    Take the mean over all non-batch dimensions.
    """
    return tensor.mean(dim=list(range(1, len(tensor.shape))))


def normalization(channels):
    """
    Make a standard normalization layer.

    :param channels: number of input channels.
    :return: an nn.Module for normalization.
    """
    return nn.GroupNorm(32, channels)


def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.

    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = th.exp(
        -math.log(max_period) * th.arange(start=0, end=half, dtype=th.float32) / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = th.cat([th.cos(args), th.sin(args)], dim=-1)
    if dim % 2:
        embedding = th.cat([embedding, th.zeros_like(embedding[:, :1])], dim=-1)
    return embedding



r"""Denoisers, parametrizations and training objectives.

For a distribution :math:`p(X)` over :math:`\mathbb{R}^D` and a perturbation kernel

.. math:: p(X_t \mid X) = \mathcal{N}(X_t \mid \alpha_t X, \sigma_t^2 I) \, ,

the goal of a denoiser is to predict :math:`X` given :math:`X_t`, that is to infer the
posterior distribution

.. math:: p(X \mid X_t) = \frac{p(X) \, p(X_t \mid X)}{p(X_t)} \, .

A denoiser is therefore an approximation :math:`q_\phi(X \mid X_t)` of the posterior
:math:`p(X \mid X_t)` and its objective is to minimize the Kullback-Leibler (KL)
divergence

.. math::
    & \arg \min_\phi \mathbb{E}_{x_t \,\sim\, p(X_t)}
        \big[ \mathrm{KL}( p(X \mid x_t) \parallel q_\phi(X \mid x_t) ) \big] \\
    = \, & \arg \min_\phi \mathbb{E}_{x, x_t \,\sim\, p(X, X_t)}
        \big[ -\log q_\phi(x \mid x_t) \big] \, .

For most use cases, it is enough to estimate the mean :math:`\mathbb{E}[X \mid x_t]` of
the posterior :math:`p(X \mid x_t)`, in which case :math:`q_\phi(X \mid x_t)` should
have mean :math:`\mu_\phi(x_t) \approx \mathbb{E}[X \mid x_t]`.
"""

__all__ = [
    "Posterior",
    "DiracPosterior",
    "GaussianPosterior",
    "Denoiser",
    "GaussianDenoiser",
    "SimpleDenoiser",
    "KarrasDenoiser",
]

import abc
import math
import torch
import torch.nn as nn

from collections.abc import Callable
from torch import Tensor

from .linalg.covariance import Covariance, IsotropicCovariance
from .nn.utils import get_module_dtype
from .noise import Schedule


class Posterior(abc.ABC):
    r"""Abstract posterior :math:`q_\phi(X \mid x_t)`."""

    mean: Tensor


class DiracPosterior(Posterior):
    r"""Creates a Dirac delta posterior :math:`\delta(X - \mu)`.

    Arguments:
        mean: The mean :math:`\mu`, with shape :math:`(*)`.
    """

    mean: Tensor

    def __init__(self, mean: Tensor) -> None:
        self.mean = mean


class GaussianPosterior(Posterior):
    r"""Creates a Gaussian posterior :math:`\mathcal{N}(X \mid \mu, \sigma^2)`.

    Arguments:
        mean: The mean :math:`\mu`, with shape :math:`(*)`.
        var: The variance :math:`\sigma^2`, with shape :math:`(*)`.
    """

    mean: Tensor
    var: Tensor

    def __init__(self, mean: Tensor, var: Tensor) -> None:
        self.mean = mean
        self.var = var

    def log_prob(self, x: Tensor) -> Tensor:
        r"""
        Arguments:
            x: A tensor :math:`x`, with shape :math:`(*)`.

        Returns:
            The log-density :math:`\log \mathcal{N}(x \mid \mu, \sigma^2)`, with shape
            :math:`(*)`.
        """

        return -((x - self.mean) ** 2 / self.var + torch.log(self.var) + math.log(2 * math.pi)) / 2


class Denoiser(nn.Module):
    r"""Abstract denoiser module."""

    schedule: Schedule

    @abc.abstractmethod
    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> Posterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The posterior :math:`q_\phi(X \mid x_t)`.
        """

        pass


class GaussianDenoiser(Denoiser):
    r"""Creates an analytical Gaussian denoiser.

    Let :math:`X \sim \mathcal{N}(\mu_x, \Sigma_x)` and :math:`X_t \sim \mathcal{N}(X,
    \Sigma_t)`, then

    .. math:: X \mid X_t \sim \mathcal{N} \left( \bar{\mu}, \bar{\Sigma} \right)

    with

    .. math::
        \bar{\mu} & = \mu_x + \Sigma_x \left( \Sigma_x + \Sigma_t \right)^{-1} (X_t - \mu_x) \\
        \bar{\Sigma} & = \left( \Sigma_x^{-1} + \Sigma_t^{-1} \right)^{-1}

    References:
        | Bayesian Filtering and Smoothing (Särkkä, 2013)
        | http://www.cambridge.org/9781108926645

    Arguments:
        mean: The mean vector :math:`\mu_x`, with shape :math:`(N_1, ..., N_d)`.
        cov: The covariance matrix :math:`\Sigma_x`, with shape
            :math:`(N_1, ..., N_d, N_1, \dots, N_d)`.
    """

    def __init__(self, mean: Tensor, cov: Covariance, schedule: Schedule) -> None:
        super().__init__()

        self.mean = mean
        self.cov = cov
        self.schedule = schedule

    def _apply(self, fn: Callable[[Tensor], Tensor], recurse: bool = True) -> Denoiser:
        super()._apply(fn, recurse=recurse)

        self.mean = fn(self.mean)
        self.cov = fn(self.cov)

        return self

    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`()`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \bar{\mu})`.
        """

        alpha_t, sigma_t = self.schedule(t)

        cov_t = IsotropicCovariance((sigma_t / alpha_t) ** 2)
        cov_inv = (self.cov + cov_t).inv

        mean = self.mean + self.cov(cov_inv(x_t / alpha_t - self.mean))

        return DiracPosterior(mean=mean)


class SimpleDenoiser(Denoiser):
    r"""Creates a denoiser with simple preconditioning.

    .. math::
        \mu_\phi(x_t) = b_\phi(c_\mathrm{in}(t) \, x_t, c_\mathrm{time}(t))

    The preconditioning coefficients make the backbone independent of the noise
    schedule. Therefore, the latter can be replaced at any time.

    .. math::
        c_\mathrm{in}(t) & = \frac{1}{\sqrt{\alpha_t^2 + \sigma_t^2}} \\
        c_\mathrm{time}(t) & = \log \frac{\sigma_t}{\alpha_t}

    Arguments:
        backbone: A noise/time conditional network :math:`b_\phi(x_t, t)`.
        schedule: A noise schedule.
    """

    def __init__(self, backbone: nn.Module, schedule: Schedule) -> None:
        super().__init__()

        self.backbone = backbone
        self.schedule = schedule

    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> GaussianPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \mu_\phi(x_t))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_time = torch.log(sigma_t / alpha_t).reshape_as(t)

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            (c_in * x_t).to(dtype),
            c_time.to(dtype),
            **kwargs,
        ).to(x_t)

        mean = output

        return DiracPosterior(mean=mean)

    def loss(self, x: Tensor, t: Tensor, max_weight: float = 1e4, **kwargs) -> Tensor:
        r"""
        Arguments:
            x: A clean tensor :math:`x`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The weighted loss

            .. math:: \frac{\alpha_t^2 + \sigma_t^2}{\sigma_t^2} || \mu_\phi(x_t) - x ||^2

            where :math:`x_t \sim p(X_t \mid x)`, with shape :math:`(B, *)`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        z = torch.randn_like(x)
        x_t = alpha_t * x + sigma_t * z

        q = self(x_t, t, **kwargs)

        w_t = (alpha_t / sigma_t) ** 2 + 1
        w_t = torch.clip(w_t, max=max_weight)

        return (w_t * (q.mean - x).square()).mean()


class KarrasDenoiser(Denoiser):
    r"""Creates a Gaussian denoiser with EDM-style preconditioning.

    .. math:: \mu_\phi(x_t) = c_\mathrm{skip}(t) \, x_t +
        c_\mathrm{out}(t) \, b_\phi(c_\mathrm{in}(t) \, x_t, c_\mathrm{time}(t))

    The preconditioning coefficients are generalized to take the scale :math:`\alpha_t`
    into account.

    .. math::
        c_\mathrm{in}(t) & = \frac{1}{\sqrt{\alpha_t^2 + \sigma_t^2}} \\
        c_\mathrm{out}(t) & = \frac{\sigma_t}{\sqrt{\alpha_t^2 + \sigma_t^2}} \\
        c_\mathrm{skip}(t) & = \frac{\alpha_t}{\alpha_t^2 + \sigma_t^2} \\
        c_\mathrm{time}(t) & = \log \frac{\sigma_t}{\alpha_t}

    References:
        | Elucidating the Design Space of Diffusion-Based Generative Models (Karras et al., 2022)
        | https://arxiv.org/abs/2206.00364

    Arguments:
        backbone: A noise/time conditional network :math:`b_\phi(x_t, t)`.
        schedule: A noise schedule.
    """

    def __init__(self, backbone: nn.Module, schedule: Schedule) -> None:
        super().__init__()

        self.backbone = backbone
        self.schedule = schedule

    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        r"""
        Arguments:
            x_t: A noisy tensor :math:`x_t`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`()` or :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The Dirac delta :math:`\delta(X - \mu_\phi(x_t))`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_out = sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_skip = alpha_t / (alpha_t**2 + sigma_t**2)
        c_time = torch.log(sigma_t / alpha_t).reshape_as(t)

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            (c_in * x_t).to(dtype),
            c_time.to(dtype),
            **kwargs,
        ).to(x_t)

        mean = c_skip * x_t + c_out * output

        return DiracPosterior(mean=mean)

    def loss(self, x: Tensor, t: Tensor, **kwargs) -> Tensor:
        r"""
        Arguments:
            x: A clean tensor :math:`x`, with shape :math:`(B, *)`.
            t: The time :math:`t`, with shape :math:`(B)`.
            kwargs: Optional keyword arguments.

        Returns:
            The weighted loss

            .. math:: \frac{\alpha_t^2 + \sigma_t^2}{\sigma_t^2} || \mu_\phi(x_t) - x ||^2

            where :math:`x_t \sim p(X_t \mid x)`, with shape :math:`(B, *)`.
        """

        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        z = torch.randn_like(x)
        x_t = alpha_t * x + sigma_t * z

        q = self(x_t, t, **kwargs)

        w_t = (alpha_t / sigma_t) ** 2 + 1

        return (w_t * (q.mean - x).square()).mean()



r"""Noise schedules.

A noise schedule is a mapping from a time :math:`t \in [0, 1]` to the signal scale
:math:`\alpha_t \in \mathbb{R}_+` and the noise scale :math:`\sigma_t \in \mathbb{R}_+`
in a perturbation kernel

.. math:: p(X_t \mid X) = \mathcal{N}(X_t \mid \alpha_t X, \sigma_t^2 I)

from a "clean" random variable :math:`X` to a "noisy" random variable :math:`X_t`. The
only constraint is for the signal-to-noise (SNR) ratio :math:`\frac{\alpha_t}{\sigma_t}`
to be monotonically decreasing with respect to the time :math:`t`. Typically, the
initial signal scale :math:`\alpha_0` is set to 1 and the initial noise is small enough
(:math:`0 < \sigma_0 \ll 1`) that :math:`X_0` is almost equivalent to :math:`X`.

Note that the relation between :math:`X_s` and :math:`X_t` (:math:`0 \leq s \leq t`) is
not enforced by the noise schedule. For example,

.. math::
    Z & \sim \mathcal{N}(0, I) \\
    X_s & = \alpha_s X + \sigma_s Z \\
    X_t & = \alpha_t X + \sigma_t Z

and

.. math::
    Z_1, Z_2 & \sim \mathcal{N}(0, I) \\
    X_s & = \alpha_s X + \sigma_s Z_1 \\
    X_t & = \frac{\alpha_t}{\alpha_s} X_s + \sqrt{\sigma_t^2 - \frac{\alpha_t^2}{\alpha_s^2} \sigma_s^2} \, Z_2

are both compatible with the perturbation kernel :math:`p(X_t \mid X)`.
"""

__all__ = [
    "Schedule",
    "VESchedule",
    "VPSchedule",
    "CosineSchedule",
    "RectifiedSchedule",
    "DecaySchedule",
]

import abc
import math
import torch

from torch import Tensor


class Schedule(abc.ABC):
    r"""Abstract noise schedule."""

    @abc.abstractmethod
    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        r"""
        Arguments:
            t: The time :math:`t`, with shape :math:`(*)`.

        Returns:
            The signal and noise scales :math:`\alpha_t` and :math:`\sigma_t`, with
            shape :math:`(*)`.
        """

        pass


class VESchedule(Schedule):
    r"""Creates a variance exploding (VE) noise schedule.

    .. math::
        \alpha_t & = 1 \\
        \sigma_t & = \exp \big( (1 - t) \log \sigma_\min + t \log \sigma_\max \big)

    References:
        | Generative Modeling by Estimating Gradients of the Data Distribution (Song et al., 2019)
        | https://arxiv.org/pdf/1907.05600

        | Score-Based Generative Modeling through Stochastic Differential Equations (Song et al., 2021)
        | https://arxiv.org/abs/2011.13456

    Arguments:
        sigma_min: The initial noise scale :math:`\sigma_\min \in \mathbb{R}_+`.
        sigma_max: The final noise scale :math:`\sigma_\max \in \mathbb{R}_+`.
    """

    def __init__(self, sigma_min: float = 1e-3, sigma_max: float = 1e3) -> None:
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)

    def alpha(self, t: Tensor) -> Tensor:
        return torch.ones_like(t)

    def sigma(self, t: Tensor) -> Tensor:
        return torch.exp((1 - t) * math.log(self.sigma_min) + t * math.log(self.sigma_max))


class VPSchedule(Schedule):
    r"""Creates a variance preserving (VP) noise schedule.

    .. math::
        \alpha_t & = \exp \big( t^2 \log \alpha_\min \big) \\
        \sigma_t & = \sqrt{ 1 - \alpha_t^2 + \sigma_\min^2}

    References:
        | Denoising Diffusion Probabilistic Models (Ho et al. 2020)
        | https://arxiv.org/abs/2006.11239

        | Score-Based Generative Modeling through Stochastic Differential Equations (Song et al., 2021)
        | https://arxiv.org/abs/2011.13456

    Arguments:
        alpha_min: The final signal scale :math:`\alpha_\min \in ]0,1[`.
        sigma_min: The initial noise scale :math:`\sigma_\min \in ]0,1[`.
    """

    def __init__(self, alpha_min: float = 1e-3, sigma_min: float = 1e-3) -> None:
        self.alpha_min = alpha_min
        self.sigma_min = sigma_min

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)

    def alpha(self, t: Tensor) -> Tensor:
        return torch.exp(math.log(self.alpha_min) * t**2)

    def sigma(self, t: Tensor) -> Tensor:
        return torch.sqrt(1 - self.alpha(t) ** 2 + self.sigma_min**2)


class CosineSchedule(Schedule):
    r"""Creates a cosine noise schedule.

    .. math::
        \alpha_t & = \cos \big( t \arccos \alpha_\min \big) \\
        \sigma_t & = \sqrt{ 1 - \alpha_t^2 + \sigma_\min^2}

    Arguments:
        alpha_min: The final signal scale :math:`\alpha_\min \in ]0,1[`.
        sigma_min: The initial noise scale :math:`\sigma_\min \in ]0,1[`.
    """

    def __init__(self, alpha_min: float = 1e-3, sigma_min: float = 1e-3) -> None:
        self.alpha_min = alpha_min
        self.sigma_min = sigma_min

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)

    def alpha(self, t: Tensor) -> Tensor:
        return torch.cos(math.acos(self.alpha_min) * t)

    def sigma(self, t: Tensor) -> Tensor:
        return torch.sqrt(1 - self.alpha(t) ** 2 + self.sigma_min**2)


class RectifiedSchedule(Schedule):
    r"""Creates a rectified noise schedule.

    .. math::
        \alpha_t & = t \, \alpha_\min + (1 - t) \\
        \sigma_t & = t + (1 - t) \, \sigma_\min

    References:
        | Flow Straight and Fast: Learning to Generate and Transfer Data with Rectified Flow (Liu et al. 2022)
        | https://arxiv.org/abs/2209.03003

        | Flow Matching for Generative Modeling (Lipman et al. 2023)
        | https://arxiv.org/abs/2210.02747

    Arguments:
        alpha_min: The final signal scale :math:`\alpha_\min \in ]0,1[`.
        sigma_min: The initial noise scale :math:`\sigma_\min \in ]0,1[`.
    """

    def __init__(self, alpha_min: float = 1e-3, sigma_min: float = 1e-3) -> None:
        self.alpha_min = alpha_min
        self.sigma_min = sigma_min

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)

    def alpha(self, t: Tensor) -> Tensor:
        return t * self.alpha_min + (1 - t)

    def sigma(self, t: Tensor) -> Tensor:
        return t + (1 - t) * self.sigma_min


class DecaySchedule(Schedule):
    r"""Creates an exponential decay schedule.

    .. math::
        \alpha_t & = \tau \, \alpha_\min + (1 - \tau) \\
        \sigma_t & = \tau + (1 - \tau) \, \sigma_\min

    where

    .. math:: \tau = \frac{1 - \gamma^t}{1 - \gamma}

    The smaller :math:`\gamma`, the more time is allocated to high signal-to-noise
    ratios. When :math:`\gamma` is close to 1, the schedule becomes equivalent to the
    :class:`RectifiedSchedule`.

    Arguments:
        alpha_min: The final signal scale :math:`\alpha_\min \in ]0,1[`.
        sigma_min: The initial noise scale :math:`\sigma_\min \in ]0,1[`.
        gamma: The decay factor :math:`\gamma \in ]0,1[`.
    """

    def __init__(
        self, alpha_min: float = 1e-3, sigma_min: float = 1e-3, gamma: float = 0.1
    ) -> None:
        self.alpha_min = alpha_min
        self.sigma_min = sigma_min
        self.gamma = gamma

    def __call__(self, t: Tensor) -> tuple[Tensor, Tensor]:
        return self.alpha(t), self.sigma(t)

    def tau(self, t: Tensor) -> Tensor:
        return (1 - self.gamma**t) / (1 - self.gamma)

    def alpha(self, t: Tensor) -> Tensor:
        t = self.tau(t)
        return t * self.alpha_min + (1 - t)

    def sigma(self, t: Tensor) -> Tensor:
        t = self.tau(t)
        return t + (1 - t) * self.sigma_min



from . import cc12m_1, danbooru_128, imagenet_128, wikiart_128, wikiart_256, yfcc_1, yfcc_2

models = {
    "cc12m_1": cc12m_1.CC12M1Model,
    "cc12m_1_cfg": cc12m_1.CC12M1Model,
    "danbooru_128": danbooru_128.Danbooru128Model,
    "imagenet_128": imagenet_128.ImageNet128Model,
    "wikiart_128": wikiart_128.WikiArt128Model,
    "wikiart_256": wikiart_256.WikiArt256Model,
    "yfcc_1": yfcc_1.YFCC1Model,
    "yfcc_2": yfcc_2.YFCC2Model,
}


def get_model(model):
    return models[model]


def get_models():
    return list(models.keys())



r"""Diffusion Posterior Sampling (DPS) internals.

References:
    | Diffusion Posterior Sampling for General Noisy Inverse Problems (Chung et al., 2022)
    | https://arxiv.org/abs/2209.14687
"""

__all__ = [
    "DPSSampler",
]

import torch

from collections.abc import Callable
from torch import Tensor

from ..denoise import Denoiser
from ..sample import DDPMSampler


class DPSSampler(DDPMSampler):
    r"""Creates a DPS sampler.

    Arguments:
        denoiser: A denoiser :math:`q_\phi(X \mid X_t)`.
        y: An observation :math:`y \sim \mathcal{N}(A(x), \Sigma_y)`.
        A: The forward operator :math:`x \mapsto A(x)`.
        zeta: The guidance strength :math:`\zeta`.
        kwargs: Keyword arguments passed to :class:`azula.sample.DDPMSampler`.
    """

    def __init__(
        self,
        denoiser: Denoiser,
        y: Tensor,
        A: Callable[[Tensor], Tensor],
        zeta: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(denoiser, **kwargs)

        self.y = y
        self.A = A
        self.zeta = zeta

    @torch.no_grad()
    def step(self, x_t: Tensor, t: Tensor, s: Tensor, **kwargs) -> Tensor:
        # DDPM
        alpha_s, sigma_s = self.denoiser.schedule(s)
        alpha_t, sigma_t = self.denoiser.schedule(t)

        tau = 1 - (alpha_t / alpha_s * sigma_s / sigma_t) ** 2
        eps = torch.randn_like(x_t)

        with torch.enable_grad():
            x_t = x_t.detach().requires_grad_()
            x_hat = self.denoiser(x_t, t, **kwargs).mean

        x_s = alpha_s * x_hat
        x_s = x_s + sigma_s * torch.sqrt(1 - tau) / sigma_t * (x_t - alpha_t * x_hat)
        x_s = x_s + sigma_s * torch.sqrt(tau) * eps

        # DPS
        with torch.enable_grad():
            error = self.y - self.A(x_hat)
            norm = torch.linalg.vector_norm(error)

        grad = torch.autograd.grad(norm, x_t)[0]

        return x_s - self.zeta * grad



import math
import torch

from torch import nn


class ResidualBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return self.main(input) + self.skip(input)


class ResConvBlock(ResidualBlock):
    def __init__(self, c_in, c_mid, c_out, dropout_last=True) -> None:
        skip = None if c_in == c_out else nn.Conv2d(c_in, c_out, 1, bias=False)
        super().__init__(
            [
                nn.Conv2d(c_in, c_mid, 3, padding=1),
                nn.ReLU(inplace=True),
                nn.Conv2d(c_mid, c_out, 3, padding=1),
                nn.ReLU(inplace=True) if dropout_last else nn.Identity(),
            ],
            skip,
        )


class SkipBlock(nn.Module):
    def __init__(self, main, skip=None) -> None:
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        return torch.cat([self.main(input), self.skip(input)], dim=1)


class FourierFeatures(nn.Module):
    def __init__(self, in_features, out_features, std=1.0) -> None:
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(torch.randn([out_features // 2, in_features]) * std)

    def forward(self, input):
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)


class SelfAttention2d(nn.Module):
    def __init__(self, c_in, n_head=1, dropout_rate=0.1) -> None:
        super().__init__()
        assert c_in % n_head == 0
        self.norm = nn.GroupNorm(1, c_in)
        self.n_head = n_head
        self.qkv_proj = nn.Conv2d(c_in, c_in * 3, 1)
        self.out_proj = nn.Conv2d(c_in, c_in, 1)
        self.dropout = nn.Identity()  # nn.Dropout2d(dropout_rate, inplace=True)

    def forward(self, input):
        n, c, h, w = input.shape
        qkv = self.qkv_proj(self.norm(input))
        qkv = qkv.view([n, self.n_head * 3, c // self.n_head, h * w]).transpose(2, 3)
        q, k, v = qkv.chunk(3, dim=1)
        scale = k.shape[3] ** -0.25
        att = ((q * scale) @ (k.transpose(2, 3) * scale)).softmax(3)
        y = (att @ v).transpose(2, 3).contiguous().view([n, c, h, w])
        return input + self.dropout(self.out_proj(y))


def expand_to_planes(input, shape):
    return input[..., None, None].repeat([1, 1, shape[2], shape[3]])


class YFCC1Model(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.shape = (3, 512, 512)
        self.min_t = 0.0
        self.max_t = 1.0

        c = 128  # The base channel count

        # The inputs to timestep_embed will approximately fall into the range
        # -10 to 10, so use std 0.2 for the Fourier Features.
        self.timestep_embed = FourierFeatures(1, 16)

        self.net = nn.Sequential(  # 512x512
            ResConvBlock(3 + 16, c, c),
            ResConvBlock(c, c, c),
            ResConvBlock(c, c, c),
            ResConvBlock(c, c, c),
            SkipBlock([
                nn.AvgPool2d(2),  # 512x512 -> 256x256
                ResConvBlock(c, c, c),
                ResConvBlock(c, c, c),
                ResConvBlock(c, c, c),
                ResConvBlock(c, c, c),
                SkipBlock([
                    nn.AvgPool2d(2),  # 256x256 -> 128x128
                    ResConvBlock(c, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c * 2),
                    SkipBlock([
                        nn.AvgPool2d(2),  # 128x128 -> 64x64
                        ResConvBlock(c * 2, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        SkipBlock([
                            nn.AvgPool2d(2),  # 64x64 -> 32x32
                            ResConvBlock(c * 2, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 4),
                            SkipBlock([
                                nn.AvgPool2d(2),  # 32x32 -> 16x16
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                SkipBlock([
                                    nn.AvgPool2d(2),  # 16x16 -> 8x8
                                    ResConvBlock(c * 4, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    SkipBlock([
                                        nn.AvgPool2d(2),  # 8x8 -> 4x4
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        ResConvBlock(c * 8, c * 8, c * 8),
                                        SelfAttention2d(c * 8, c * 8 // 64),
                                        nn.Upsample(
                                            scale_factor=2, mode="bilinear", align_corners=False
                                        ),
                                    ]),
                                    ResConvBlock(c * 16, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 8),
                                    SelfAttention2d(c * 8, c * 8 // 64),
                                    ResConvBlock(c * 8, c * 8, c * 4),
                                    SelfAttention2d(c * 4, c * 4 // 64),
                                    nn.Upsample(
                                        scale_factor=2, mode="bilinear", align_corners=False
                                    ),
                                ]),
                                ResConvBlock(c * 8, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                ResConvBlock(c * 4, c * 4, c * 4),
                                SelfAttention2d(c * 4, c * 4 // 64),
                                nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False),
                            ]),
                            ResConvBlock(c * 8, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 4),
                            ResConvBlock(c * 4, c * 4, c * 2),
                            nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False),
                        ]),
                        ResConvBlock(c * 4, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        ResConvBlock(c * 2, c * 2, c * 2),
                        nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False),
                    ]),
                    ResConvBlock(c * 4, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c * 2),
                    ResConvBlock(c * 2, c * 2, c),
                    nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False),
                ]),
                ResConvBlock(c * 2, c, c),
                ResConvBlock(c, c, c),
                ResConvBlock(c, c, c),
                ResConvBlock(c, c, c),
                nn.Upsample(scale_factor=2, mode="bilinear", align_corners=False),
            ]),
            ResConvBlock(c * 2, c, c),
            ResConvBlock(c, c, c),
            ResConvBlock(c, c, c),
            ResConvBlock(c, c, 3, dropout_last=False),
        )

    def forward(self, input, t):
        timestep_embed = expand_to_planes(self.timestep_embed(t[:, None]), input.shape)
        return self.net(torch.cat([input, timestep_embed], dim=1))



r"""Velocity diffusion model (VDM) plugin.

.. code-block:: python

    from azula.plugins import vdm

References:
    | https://github.com/crowsonkb/v-diffusion-pytorch
"""

__all__ = [
    "VelocityDenoiser",
    "load_model",
]

import math
import torch
import torch.nn as nn

from torch import Tensor

from azula.denoise import Denoiser, DiracPosterior
from azula.hub import download
from azula.nn.utils import get_module_dtype, skip_init
from azula.noise import Schedule, VPSchedule

from ._src import get_model
from ..utils import load_cards


class VelocityDenoiser(Denoiser):
    r"""Creates a velocity denoiser.

    Arguments:
        backbone: A time conditional network.
        schedule: A noise schedule. If :py:`None`, use :class:`azula.noise.VPSchedule`
            instead.
    """

    def __init__(
        self,
        backbone: nn.Module,
        schedule: Schedule | None = None,
    ) -> None:
        super().__init__()

        self.backbone = backbone

        if schedule is None:
            self.schedule = VPSchedule(alpha_min=1e-2, sigma_min=1e-2)
        else:
            self.schedule = schedule

    def forward(self, x_t: Tensor, t: Tensor, **kwargs) -> DiracPosterior:
        alpha_t, sigma_t = self.schedule(t)

        while alpha_t.ndim < x_t.ndim:
            alpha_t, sigma_t = alpha_t[..., None], sigma_t[..., None]

        c_in = torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_out = -sigma_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_skip = alpha_t * torch.rsqrt(alpha_t**2 + sigma_t**2)
        c_time = torch.atan2(sigma_t, alpha_t).flatten() / math.pi * 2

        dtype = get_module_dtype(self.backbone)

        output = self.backbone(
            (c_in * x_t).to(dtype),
            c_time.to(dtype),
            **kwargs,
        ).to(x_t)

        mean = c_skip * x_t + c_out * output

        return DiracPosterior(mean=mean)


def load_model(name: str, **kwargs) -> Denoiser:
    r"""Loads a pre-trained VDM denoiser.

    Arguments:
        name: The pre-trained model name.
        kwargs: Keyword arguments passed to :func:`torch.load`.

    Returns:
        A pre-trained denoiser.
    """

    kwargs.setdefault("map_location", "cpu")
    kwargs.setdefault("weights_only", True)

    card = load_cards(__name__)[name]
    state = torch.load(download(card.url, hash_prefix=card.hash), **kwargs)

    with skip_init():
        denoiser = make_model(**card.config)

    denoiser.backbone.load_state_dict(state)

    return denoiser.eval()


def make_model(
    model: str = "imagenet_128",
) -> Denoiser:
    r"""Initializes a VDM denoiser."""

    backbone = get_model(model)()

    return VelocityDenoiser(backbone)



r"""Linear system solvers."""

__all__ = [
    "cg",
    "gmres",
]

import torch

from collections.abc import Callable
from torch import Tensor


def cg(
    A: Callable[[Tensor], Tensor],
    b: Tensor,
    x0: Tensor | None = None,
    iterations: int = 1,
    dtype: torch.dtype | None = None,
) -> Tensor:
    r"""Solves a linear system :math:`Ax = b` with conjugate gradient (CG) iterations.

    The matrix :math:`A \in \mathbb{R}^{D \times D}` must be symmetric positive
    (semi)definite.

    Wikipedia:
        https://wikipedia.org/wiki/Conjugate_gradient_method

    Warning:
        This function is optimized for GPU execution. To avoid CPU-GPU communication,
        all iterations are performed regardless of convergence.

    Arguments:
        A: The linear operator :math:`x \mapsto Ax`.
        b: The right-hand side vector :math:`b`, with shape :math:`(*, D)`.
        x0: An initial guess :math:`x_0`, with shape :math:`(*, D)`. If :py:`None`, use
            :math:`x_0 = 0` instead.
        iterations: The number of CG iterations :math:`n`.
        dtype: The data type used for intermediate computations. If :py:`None`, use
            :class:`torch.float64` instead.

    Returns:
        The :math:`n`-th iteration :math:`x_n`, with shape :math:`(*, D)`.
    """

    if dtype is None:
        dtype = torch.float64

    epsilon = torch.finfo(dtype).smallest_normal

    if x0 is None:
        x = torch.zeros_like(b)
        r = b
    else:
        x = x0
        r = b - A(x0)

    x = x.to(dtype)
    r = r.to(dtype)
    rr = torch.einsum("...i,...i", r, r)
    p = r

    for _ in range(iterations):
        Ap = A(p.to(b)).to(dtype)
        pAp = torch.einsum("...i,...i", p, Ap)
        alpha = rr / torch.clip(pAp, min=epsilon)
        x_ = x + alpha[..., None] * p
        r_ = r - alpha[..., None] * Ap
        rr_ = torch.einsum("...i,...i", r_, r_)
        beta = rr_ / torch.clip(rr, min=epsilon)
        p_ = r_ + beta[..., None] * p

        x, r, rr, p = x_, r_, rr_, p_

    return x.to(b)


def gmres(
    A: Callable[[Tensor], Tensor],
    b: Tensor,
    x0: Tensor | None = None,
    iterations: int = 1,
    dtype: torch.dtype | None = None,
) -> Tensor:
    r"""Solves a linear system :math:`Ax = b` with generalized minimal residual (GMRES) iterations.

    The matrix :math:`A \in \mathbb{R}^{D \times D}` can be non-symmetric non-definite.

    Wikipedia:
        https://wikipedia.org/wiki/Generalized_minimal_residual_method

    Warning:
        This function is optimized for GPU execution. To avoid CPU-GPU communication,
        all iterations are performed regardless of convergence.

    Arguments:
        A: The linear operator :math:`x \mapsto Ax`.
        b: The right-hand side vector :math:`b`, with shape :math:`(*, D)`.
        x0: An initial guess :math:`x_0`, with shape :math:`(*, D)`. If :py:`None`, use
            :math:`x_0 = 0` instead.
        iterations: The number of GMRES iterations :math:`n`.
        dtype: The data type used for intermediate computations. If :py:`None`, use
            :class:`torch.float64` instead.

    Returns:
        The :math:`n`-th iteration :math:`x_n`, with shape :math:`(*, D)`.
    """

    if dtype is None:
        dtype = torch.float64

    epsilon = torch.finfo(dtype).smallest_normal

    if x0 is None:
        r = b
    else:
        r = b - A(x0)

    r = r.to(dtype)

    def normalize(x: Tensor) -> tuple[Tensor, Tensor]:
        norm = torch.linalg.vector_norm(x, dim=-1)
        x = x / torch.clip(norm[..., None], min=epsilon)

        return x, norm

    def rotation(a: Tensor, b: Tensor) -> tuple[Tensor, Tensor]:
        c = torch.clip(torch.sqrt(a * a + b * b), min=epsilon)
        return a / c, -b / c

    V = [None for _ in range(iterations + 1)]
    B = [None for _ in range(iterations + 1)]
    H = [[None for _ in range(iterations)] for _ in range(iterations + 1)]
    cs = [None for _ in range(iterations)]
    ss = [None for _ in range(iterations)]

    V[0], B[0] = normalize(r)

    for j in range(iterations):
        v = V[j].to(b)
        w = A(v).to(dtype)

        # Apply Arnoldi iteration to get the j+1-th basis
        for i in range(j + 1):
            H[i][j] = torch.einsum("...i,...i", w, V[i])
            w = w - H[i][j][..., None] * V[i]
        w, w_norm = normalize(w)
        H[j + 1][j] = w_norm
        V[j + 1] = w

        # Apply Givens rotation
        for i in range(j):
            tmp = cs[i] * H[i][j] - ss[i] * H[i + 1][j]
            H[i + 1][j] = cs[i] * H[i + 1][j] + ss[i] * H[i][j]
            H[i][j] = tmp

        cs[j], ss[j] = rotation(H[j][j], H[j + 1][j])
        H[j][j] = cs[j] * H[j][j] - ss[j] * H[j + 1][j]

        # Update residual vector
        B[j + 1] = ss[j] * B[j]
        B[j] = cs[j] * B[j]

        # Fill with zeros
        for i in range(j + 1, iterations + 1):
            H[i][j] = torch.zeros_like(H[j][j])

    V, B, H = V[:-1], B[:-1], H[:-1]

    V = torch.stack(V, dim=-2)
    B = torch.stack(B, dim=-1)
    H = torch.stack([torch.stack(Hi, dim=-1) for Hi in H], dim=-2)

    y = torch.linalg.solve_triangular(
        H + epsilon * torch.eye(iterations, dtype=dtype, device=H.device),
        B.unsqueeze(dim=-1),
        upper=True,
    ).squeeze(dim=-1)

    if x0 is None:
        x = torch.einsum("...ij,...i", V, y)
    else:
        x = x0 + torch.einsum("...ij,...i", V, y)

    return x.to(b)
